/**
  *  \file client/tiles/planetmining.cc
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/tiles/planetmining.h"
#include "cpluslib/format.h"
#include "game/planet.h"
#include "game/tables.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/misc.h"
#include "util/nls.h"
#include "ui/table.h"
#include "game/global.h"
#include "game/format.h"
#include "cpluslib/misc.h"
#include "util/strtab.h"
#include "resmgr/resmgr.h"

WPlanetMiningTile::WPlanetMiningTile(GPlanet& planet, GCargoType ele)
    : WTile(GfxPoint(295, 5 * font_heights[FONT_NORMAL] + 11 /* 3x 3px line, 2x 1px frame */)),
      planet(planet), ele(ele),
      mines(planet.getStructures(str_Mines))
{ }

static void
drawPercentBar(GfxCanvas& can, GfxRect rect, long fill, int divi, int color)
{
    int split = rect.w * fill / (100*divi);
    if (fill && !split)
        split = 1;
    if (split > rect.w)
        split = rect.w;

    drawSolidBar(can, GfxRect(rect.x, rect.y, split, rect.h), standard_colors[color]);
    drawTiledArea(can, GfxRect(rect.x+split, rect.y, rect.w-split, rect.h),
                  requestPixmap("bluetile"), standard_colors[COLOR_SHIELD+3], 0);
}

void
WPlanetMiningTile::drawData(GfxCanvas& can)
{
    static const char*const labels[] = {
        N_("mined"),
        N_("ground"),
        N_("density"),
        N_("total"),
    };
    static const char*const units[] = {
        " kt",
        " %",
    };

    /* Frame around the widget */
    drawFrameDown(can, getInnerSize(0, 0));

    /* Geometry */
    const GfxRect in = getInnerSize(5, 5);
    const int labelWidth = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(labels, countof(labels))) + 5;
    const int unitWidth = getTextWidth(*fonts[FONT_NORMAL], StaticStringTable(units, countof(units))) + 5;
    const int amountWidth = fonts[FONT_NORMAL]->getTextWidth("0") * 5 + 5;
    const int barWidth = std::max(int(in.w - labelWidth - unitWidth - amountWidth), 0);
    const int labelX = in.x;
    const int barX = labelX + labelWidth;
    const int uaX = barX + barWidth + amountWidth;
    const int he = font_heights[FONT_NORMAL];
    int y = in.y + 3;

    /* Canvas */
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Heading);
    ctx.setSolidBackground();

    /* Mineral type */
    outTextF(ctx, 5, y, in.w, getCargoName(ele));
    y += he;

    /* Age */
    bool haveAge = false;
    if (ele == el_Neutronium && !planet.isPlayed() && planet.getHistoryTimestamp(ts_Minerals) < getCurrentTurn().getTurnNumber()) {
        /* Planet is not current. Show age only for Neutronium (first mineral) */
        ctx.setTextAlign(2, 0);
        ctx.setColor(UIColor::tc_Yellow);
        outText(ctx, in.w+5, 3, formatTurnNumber(planet.getHistoryTimestamp(ts_Minerals)));
        ctx.setTextAlign(0, 0);
        haveAge = true;
    }

    ctx.setColor(UIColor::tc_Static);
    if (planet.getOreGround(ele).isKnown()
        && planet.getOreDensity(ele).isKnown()
        && planet.isCargoAmountKnown(ele))
    {
        int line2Reserve = 0;

        /* Mining */
        if (mines.isKnown() && !haveAge) {
            /* Planet is current, and we know its number of mines */
            int amount = planet.getMiningCapacity(ele, mines);
            int32_t ground = planet.getOreGround(ele);

            // Show mining capacity
            if (amount > ground) {
                ctx.setColor(UIColor::tc_Yellow);
            } else {
                ctx.setColor(UIColor::tc_Green);
            }
            ctx.setTextAlign(2, 0);
            outText(ctx, in.w+5, 3, format(_("%d kt/turn"), amount));

            // Show mining time
            string_t timestr;
            if (amount != 0) {
                int32 time = (ground + (amount-1)) / amount;
                if (time > 30)
                    timestr = _(">30 turns");
                else
                    timestr = format(_("\xE2\x89\x88" "%d turn%!1{s%}"), time);
            }
            line2Reserve = fonts[FONT_NORMAL]->getTextWidth(timestr);
            outText(ctx, in.w+5, y, timestr);
            ctx.setTextAlign(0, 0);
        }

        /* Ore Summary */
        int line2 = in.w - line2Reserve - 5;
        if (line2 > 0) {
            ctx.setColor(UIColor::tc_Static);
            if (planet.getOreGround(ele)) {
                outTextF(ctx, 5, y, line2,
                         format("%s, %s",
                                getMineralMassClassName(planet.getOreGround(ele)),
                                getMineralDensityClassName(planet.getOreDensity(ele))));
            } else {
                outTextF(ctx, 5, y, line2, _("none"));
            }
        }
        y += he;
        y += 3;

        /* Bars */
        outText(ctx, in.x, y, _(labels[0]));
        drawPercentBar(can, GfxRect(barX, y, barWidth, he-1), planet.getCargo(ele), 82, COLOR_WHITE);
        ctx.setTextAlign(2, 0);
        outTextF(ctx, uaX, y, amountWidth, numToString(planet.getCargo(ele)));
        ctx.setTextAlign(0, 0);
        outText(ctx, uaX, y, units[0]);
        y += he;

        outText(ctx, in.x, y, _(labels[1]));
        drawPercentBar(can, GfxRect(barX, y, barWidth, he-1), planet.getOreGround(ele), 82, COLOR_WHITE);
        ctx.setTextAlign(2, 0);
        outTextF(ctx, uaX, y, amountWidth, numToString(planet.getOreGround(ele)));
        ctx.setTextAlign(0, 0);
        outText(ctx, uaX, y, units[0]);
        y += he;

        outText(ctx, in.x, y, _(labels[2]));
        drawPercentBar(can, GfxRect(barX, y, barWidth, he-1), planet.getOreDensity(ele), 1, COLOR_GRAY);
        ctx.setTextAlign(2, 0);
        outTextF(ctx, uaX, y, amountWidth, numToString(planet.getOreDensity(ele)));
        ctx.setTextAlign(0, 0);
        outText(ctx, uaX, y, units[1]);
        y += he;
    } else {
        int32_t n = -1;
        if (planet.isCargoAmountKnown(ele)) {
            n = planet.getCargo(ele);
        } else if (planet.getOreGround(ele).isKnown()) {
            n = planet.getOreGround(ele);
        }

        if (n >= 0) {
            if (n != 0) {
                outTextF(ctx, 5, y, in.w, getMineralMassClassName(n));
            } else {
                outTextF(ctx, 5, y, in.w, _("none"));
            }
            y += he;
            y += 3;

            outText(ctx, in.x, y, _(labels[3]));
            drawPercentBar(can, GfxRect(barX, y, barWidth, he-1), n, 82, COLOR_WHITE);
            ctx.setTextAlign(2, 0);
            outTextF(ctx, uaX, y, amountWidth, numToString(n));
            ctx.setTextAlign(0, 0);
            outText(ctx, uaX, y, units[0]);
            y += he;
            y += 3;

            outTextFormatted(ctx, in.x + 10, y, in.w - 10, _("(no information on density and mined/ground masses available.)"));
        }
    }
}

void
WPlanetMiningTile::setNumMines(mp16_t mines)
{
    this->mines = mines;
    drawWidget();
}




WNarrowPlanetMineralTile::WNarrowPlanetMineralTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(250, 5 * font_heights[FONT_NORMAL]), type)
{ }

void
WNarrowPlanetMineralTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());

    GPlanet* pl = dynamic_cast<GPlanet*>(getObject());
    if (!pl) {
        drawBackground(ctx, getInnerSize(0, 0));
        return;
    }

    enum { N = 4 };
    static const GCargoType eles[N] = { el_Neutronium, el_Tritanium, el_Duranium, el_Molybdenum };

    UITableWriter tw(ctx, getInnerSize(0, 0), FONT_NORMAL);
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    tw.writeFixedCell(UIColor::tc_Static, _("Neu:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Tri:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Dur:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Mol:"));

    /* Mined */
    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("9,999,999"));
    tw.writeFixedCell(UIColor::tc_Static, _("Mined"));
    for (int i = 0; i < N; ++i) {
        if (pl->isCargoAmountKnown(eles[i])) {
            tw.writeVariableCell(UIColor::tc_Green, numToString(pl->getCargoRaw(eles[i])));
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }
    tw.newLColumn(fonts[FONT_NORMAL]->getTextWidth("kt"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    for (int i = 0; i < N; ++i) {
        if (pl->isCargoAmountKnown(eles[i])) {
            tw.writeVariableCell(UIColor::tc_Green, "kt");
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }

    /* Ground */
    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("9,999,999"));
    tw.writeFixedCell(UIColor::tc_Static, _("Ground"));
    for (int i = 0; i < N; ++i) {
        if (pl->getOreGround(eles[i]).isKnown()) {
            tw.writeVariableCell(UIColor::tc_Green, numToString(pl->getOreGround(eles[i])));
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }
    tw.newLColumn(fonts[FONT_NORMAL]->getTextWidth("kt"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    for (int i = 0; i < N; ++i) {
        if (pl->getOreGround(eles[i]).isKnown()) {
            tw.writeVariableCell(UIColor::tc_Green, "kt");
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }

    /* Density */
    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("100%"));
    tw.writeFixedCell(UIColor::tc_Static, string_t());
    for (int i = 0; i < N; ++i) {
        if (pl->getOreDensity(eles[i]).isKnown()) {
            tw.writeVariableCell(UIColor::tc_Green, numToString(pl->getOreDensity(eles[i])) + "%");
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }
}
