/**
  *  \file client/widgets/racelist.cc
  *  \brief Race List widget
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/racelist.h"
#include "gfx/gfxutil.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "game/racename.h"
#include "game/team.h"
#include "game/player.h"

/** Color mapping. Indexed by player-number-minus-one. In PCC 1.x, the statistic
    and combat diagram screens have a separate palette for races 7-11. Since we
    usually have close colors, we use those instead. */
const uint8 WRaceList::colors[] = {
    COLOR_BLACK,          /* Fed: black */
    COLOR_WHITE,          /* Lizard: white */
    COLOR_YELLOW,         /* Bird: yellow */
    COLOR_GREEN,          /* Klingon: green */
    COLOR_BLUE,           /* Pirate: blue */
    COLOR_RED,            /* Cyborg: red */
    COLOR_BRIGHTMAGENTA,  /* Crystal: lilac. 63/24/63 [255/97/255] in PCC 1.x, 255/85/255 here */
    COLOR_BRIGHTBROWN,    /* Empire: brown. 48/24/0 [194/97/255] in PCC 1.x, 194/97/255 here */
    COLOR_BRIGHTORANGE,   /* Robot: orange. 63/48/0 [255/194/0] in PCC 1.x, 255/194/0 here */
    COLOR_BRIGHTCYAN,     /* Rebel: cyan. 32/63/63 [128/255/255] in PCC 1.x, 85/255/255 here */
    COLOR_DARKGREEN,      /* Colony: green. 0/48/0 [0/194/0] in PCC 1.x, 0/170/0 here */
    COLOR_BLUEBLACK       /* Aliens: blue/gray */
};

/** Constructor.
    \param preferred_width preferred width of widget, in pixels. For FlowLayout: maximum width.
                           For VerticalLayout: minimum width. For HorizontalLayout: ignored.
    \param flags           Layout flags
    \param from            first player to display, [1,NUM_OWNERS]
    \param to              last player to display, inclusive, [from,NUM_OWNERS] */
WRaceList::WRaceList(int preferred_width, int flags, int from, int to)
    : preferred_width(preferred_width),
      from(from),
      to(to),
      current_click(0),
      flags(flags)
{ }

/** Destructor. */
WRaceList::~WRaceList()
{ }

/** Compute layout of widget.
    \param positions [out] Relative positions of player boxes
    \param w [in] Assumed widget width in pixels */
void
WRaceList::calcLayout(GPlayerArray<GfxRect>& positions, int w)
{
    int x = 0, y = 0;
    int cellheight = font_heights[FONT_NORMAL];

    // Compute cell width
    GPlayerArray<int> cellwidths;
    for (int i = from; i <= to; ++i) {
        if (flags & ShowLetters)
            cellwidths[i] = cellheight;
        else if (flags & ShowTeamNames)
            cellwidths[i] = fonts[FONT_NORMAL]->getTextWidth(getTeamName(i));
        else
            cellwidths[i] = fonts[FONT_NORMAL]->getTextWidth(player_racenames.getAdjName(i));
    }

    // Compute layout
    if (flags & VerticalLayout) {
        // VerticalLayout. Find maximum width:
        int maxwidth = preferred_width;
        for (int i = from; i <= to; ++i)
            if (maxwidth < cellwidths[i])
                maxwidth = cellwidths[i];
        // Assign positions:
        for (int i = from; i <= to; ++i) {
            positions[i] = GfxRect(x, y, maxwidth, cellheight);
            y += cellheight;
        }
    } else {
        // FlowLayout or HorizontalLayout
        for (int i = from; i <= to; ++i) {
            int cellwidth = cellwidths[i];
            if (cellwidth > 0) {
                /* cell with content */
                if ((flags & FlowLayout) != 0 && x + cellwidth > w) {
                    /* start new line */
                    x = 0;
                    y += cellheight;
                }
                positions[i] = GfxRect(x, y, cellwidth, cellheight);
                x += cellwidth;
                if (!(flags & ShowLetters))
                    x += font_ems[FONT_NORMAL]/2;
            } else {
                /* empty cell */
                positions[i] = GfxRect(x, y, 0, cellheight);
            }
        }
    }
}

void
WRaceList::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, getExtent());
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);
    calcLayout(positions, extent.w);
    for (int i = from; i <= to; ++i) {
        GfxPoint pos = positions[i].getTopLeft().movedBy(extent.getTopLeft());
        // Set color
        if (highlight_races.contains(i))
            ctx.setColor(COLOR_GREEN);
        else if (flags & NoColors)
            ctx.setColor(COLOR_BLACK);
        else
            ctx.setColor(colors[i-1]);

        // Draw text
        if (flags & ShowLetters)
            outText(ctx, pos.x, pos.y, string_t(1, getPlayerChar(i)));
        else if (flags & ShowTeamNames)
            outText(ctx, pos.x, pos.y, getTeamName(i));
        else
            outText(ctx, pos.x, pos.y, player_racenames.getAdjName(i));
    }
    can.update(extent);
}

void
WRaceList::getLayoutInfo(LayoutInfo& info)
{
    GPlayerArray<GfxRect> pos;
    calcLayout(pos, preferred_width);

    if (flags & FlowLayout) {
        info.pref_size = info.min_size = GfxPoint(preferred_width, pos[to].getBottomY());
        info.flags = 0;
    } else {
        info.pref_size = info.min_size = GfxPoint(pos[to].getRightX(), pos[to].getBottomY());
        if (flags & VerticalLayout)
            info.flags |= li_FlexibleH;
        else
            info.flags = 0;
    }
}

bool
WRaceList::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse) {
        if (isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
            if (e.mouse.b != 0) {
                /* mouse clicked into this widget */
                int where = 0;
                GfxPoint rel(e.mouse.x - extent.x,
                             e.mouse.y - extent.y);
                for (int i = from; i <= to; ++i) {
                    if (positions[i].contains(rel)) {
                        where = i;
                        break;
                    }
                }
                if (where != current_click && where != 0)
                    sig_click.raise(where);
                current_click = where;
                return true;
            } else {
                current_click = 0;
                return true;
            }
        } else {
            current_click = 0;
            return UIWidget::handleEvent(e, second_pass);
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
WRaceList::setHighlight(GPlayerSet set)
{
    if (set != highlight_races) {
        highlight_races = set;
        drawWidget();
    }
}

void
WRaceList::setFlags(int i)
{
    if (i != flags) {
        flags = i;
        drawWidget();
    }
}
