/**
  *  \file flak-gen.cc
  *  \brief FLAK fight generation parts
  *
  *  This module contains functions that are needed when setting up a
  *  fight. They are not needed when just playing one.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include <vector>
#include "cpluslib/assert.h"
#include "cpluslib/smallset.h"
#include "flak/flak-port.h"
#include "flak/flak.h"
#include "flak/flakconf.h"
#include "game/formula.h"
#include "cpluslib/misc.h"

/** Remove all passive objects. Units which are not attackable should
    not appear in the data to not leak too much information. */
void
FlakBattle::removePassiveObjects()
{
    // temporary array
    std::vector<int> new_ids(ships.size());

    // first, mark all referenced ships. We use the fact that a ship
    // which is in no attack list can itself not attack anyone and
    // therefore is useless. After this loop, all units which are used
    // have a 1 in new_ids[].
    for (fleet_t i = 0; i < fleets.size(); ++i) {
        FlakFleet& f = *fleets[i];
        for (int n = 0; n < f.data.att_list_size; ++n)
            if (attack_list[2 * (f.data.att_list_pointer + n) + 1] > 0)
                new_ids[attack_list[2 * (f.data.att_list_pointer + n)]] = 1;
    }

    // now, recompute stuff
    ship_t id = 0;
    for (ship_t i = 0; i < ships.size(); ++i) {
        if (new_ids[i] == 0) {
            /* ship does not appear in lists */
            new_ids[i] = -1;
            ships[i]->fleet_link->data.num_ships--;
            delete ships[i];
            ships[i] = 0;
        } else {
            new_ids[i] = id;
            ships[id] = ships[i];
            ++id;
        }
    }
    
    ships.erase(ships.begin() + id, ships.end());

    /* remap data: attack lists */
//     for (std::size_t i = 0; i < attack_list.size(); i += 2) {
//         ASSERT(new_ids[attack_list[i]] >= 0);
//         attack_list[i] = new_ids[attack_list[i]];
//     }
    std::vector<int16> new_att_list;
    for (std::size_t i = 0; i < fleets.size(); ++i) {
        FlakFleet& me = *fleets[i];
        uint32 new_start = new_att_list.size() / 2;
        uint32 new_length = 0;
        for (int n = 0; n < me.data.att_list_size; ++n) {
            int x = new_ids[attack_list[2 * (me.data.att_list_pointer + n)]];
            if (x >= 0) {
                new_att_list.push_back(x);
                new_att_list.push_back(attack_list[2 * (me.data.att_list_pointer + n) + 1]);
                ++new_length;
            }
        }
        me.data.att_list_pointer = new_start;
        me.data.att_list_size = new_length;
    }
    new_att_list.swap(attack_list);

    /* remap data: fleet pointers */
    std::size_t fleet_id = 0;
    for (std::size_t i = 0; i < fleets.size(); ++i) {
        if (fleets[i]->data.num_ships == 0) {
            /* fleet is gone */
            delete fleets[i];
            fleets[i] = 0;
        } else {
            /* fleet still there, figure out new first ship (count
               is already set up). */
            int ship_nr = fleets[i]->data.first_ship;
            while (new_ids[ship_nr] < 0)
                ++ship_nr;
            fleets[i]->data.first_ship = new_ids[ship_nr];
            fleets[fleet_id] = fleets[i];
            ++fleet_id;
        }
    }
    fleets.erase(fleets.begin() + fleet_id, fleets.end());
}

/** Compute speeds of all fleets. This function needs access to the
    host-side configuration. */
void
FlakBattle::computeFleetSpeeds()
{
    for (fleet_t i = 0; i < fleets.size(); ++i) {
        int speed = 0;
        for (int j = 0; j < fleets[i]->data.num_ships; ++j) {
            int this_speed;
            FlakShip& this_ship = *ships[j + fleets[i]->data.first_ship];
            if (this_ship.isPlanet())
                this_speed = 0;
            else
                this_speed = CONFIG_ARRAY(ShipMovementSpeed, this_ship.data.player);

            if (j == 0 || this_speed < speed)
                speed = this_speed;
        }
        fleets[i]->data.speed = speed;
    }
}

/** Compute initial fleet positions. This needs access to the
    host-side configuration. */
void
FlakBattle::computeInitialPositions()
{
    /* do we have a planet? */
    int planet_owner = 0;
    int num_players = 0;
    SmallSet<int> players;
    for (ship_t i = 0; i < ships.size(); ++i) {
        ASSERT(ships[i]->data.player > 0 && ships[i]->data.player <= NUM_OWNERS);
        if (ships[i]->isPlanet()) {
            ASSERT(!planet_owner);
            planet_owner = ships[i]->data.player;
        }
        if (i == 0 || ships[i]->data.player != ships[i-1]->data.player) {
            ASSERT(!players.contains(ships[i]->data.player));
            ++num_players;
            players |= ships[i]->data.player;
        }
    }

    /* shuffle players */
    int counter;
    int offs = flak_config.StartingDistancePerPlayer * num_players;
    int player_list[NUM_OWNERS];

    counter = 0;
    for (int i = 1; i <= NUM_OWNERS; ++i)
        if (players.contains(i))
            player_list[counter++] = i;
    ASSERT(counter == num_players);
    std::random_shuffle(player_list, player_list + num_players);

    /* assign positions */
    counter = 0;
    for (int index = 0; index < num_players; ++index) {
        const int player = player_list[index];
        if (planet_owner) {
            if (player == planet_owner) {
                assignInitialPositions(player, 0, flak_config.StartingDistancePlanet + offs);
            } else {
                assignInitialPositions(player, (PI/2) + (2*counter+1) * (PI/2) / (num_players-1),
                                       flak_config.StartingDistanceShip + offs);
                ++counter;
            }
        } else {
            assignInitialPositions(player, (PI/2) + (2*counter+1)*PI / num_players,
                                   flak_config.StartingDistanceShip + offs);
            ++counter;
        }
    }
}

/** Assign initial positions for a player.
    \param player   player number
    \param angle    angle (in radians) of first fleet
    \param dist     distance from origin, for first fleet */
void
FlakBattle::assignInitialPositions(const int player, double angle, int32 dist)
{
    int offset = flak_config.StartingDistancePerFleet;
    for (fleet_t i = 0; i < fleets.size(); ++i) {
        FlakFleet& this_fleet = *fleets[i];
        if (this_fleet.data.player != player)
            continue;
        if (this_fleet.data.speed > offset)
            offset = this_fleet.data.speed;

        this_fleet.data.x_init = roundToInt(std::cos(angle) * dist);
        this_fleet.data.y_init = roundToInt(std::sin(angle) * dist);
        dist += offset;
        angle += PI/180.0;  /* 1 degree */
    }
}

/** Adjust compensation. */
void
FlakBattle::adjustStrengths(int adj_to)
{
    int32 total_strength = 0;
    for (ship_t i = 0; i < ships.size(); ++i)
        total_strength += ships[i]->data.compensation_rating;

    int32 target_strength = ships.size() * adj_to;
    if (total_strength != 0 && total_strength < target_strength) {
        for (ship_t i = 0; i < ships.size(); ++i) {
            uint32 new_val = ships[i]->data.compensation_rating * target_strength / total_strength;
            if (new_val > flak_config.CompensationLimit)
                new_val = flak_config.CompensationLimit;
            // FIXME: this is a hack. It's well-defined but ugly.
            const_cast<TFlakShip&>(ships[i]->data).compensation_rating = new_val;
        }
    }
}

/** Initialisation, main entry. After having set up a fight, call this
    routine to compute the derived information and remove units which
    are not needed for the fight. */
void
FlakBattle::initAfterSetup()
{
    removePassiveObjects();
    computeFleetSpeeds();
    computeInitialPositions();
    if (flak_config.CompensationAdjust)
        adjustStrengths(flak_config.CompensationAdjust);
}

/** Find a possible captor for victim.
    \param victim Died ship
    \return captor, 0 if none found */
const FlakShip*
FlakBattle::findCaptor(const FlakShip& victim)
{
    /* captor still alive? */
    ASSERT(victim.status.last_hit_by);
    if (victim.status.last_hit_by->isAlive())
        return victim.status.last_hit_by;

    /* count all ships */
    int counts[FLAK_NUM_OWNERS+1];
    int total = 0;
    for (unsigned i = 0; i < countof(counts); ++i)
        counts[i] = 0;
    for (int i = 0; i < getNumShips(); ++i)
        if (getShipByNumber(i).isAlive())
            ++counts[getShipByNumber(i).data.player], ++total;

    /* captor race still alive? */
    int player = victim.status.last_hit_by->data.player;
    if (counts[player]) {
        int pick = random(counts[player]);
        for (int i = 0; i < getNumShips(); ++i)
            if (getShipByNumber(i).isAlive() && getShipByNumber(i).data.player == player)
                if (--pick < 0)
                    return &getShipByNumber(i);
    }

    /* owner race still alive? */
    player = victim.data.player;
    if (counts[player]) {
        int pick = random(counts[player]);
        for (int i = 0; i < getNumShips(); ++i)
            if (getShipByNumber(i).isAlive() && getShipByNumber(i).data.player == player)
                if (--pick < 0)
                    return &getShipByNumber(i);
    }

    /* anyone else alive? */
    if (total) {
        int pick = random(total);
        for (int i = 0; i < getNumShips(); ++i)
            if (getShipByNumber(i).isAlive())
                if (--pick < 0)
                    return &getShipByNumber(i);
    }

    /* nobody hearing me? */
    return 0;
}

/** Compute targeting rating. */
static int32
computeRating(const TFlakShip& ship)
{
    return ship.mass * flak_config.RatingMassScale
        + ship.torp_lcount * ship.torp_type * flak_config.RatingTorpScale
        + ship.beam_count * ship.beam_type * flak_config.RatingBeamScale
        + ship.bay_count * flak_config.RatingBayScale;
}

/** Compute compensation rating. */
static int
computeStrength(const TFlakShip& ship)
{
    int32 strength = flak_config.CompensationShipScale
        + ship.torp_lcount * flak_config.CompensationTorpScale
        + ship.beam_count * flak_config.CompensationBeamScale
        + ship.bay_count * flak_config.CompensationFighterScale
        + ship.mass * flak_config.CompensationMass100KTScale / 100;
    return (strength > flak_config.CompensationLimit
            ? flak_config.CompensationLimit
            : strength < 0
            ? 0
            : strength);
}

/** Initialize derived values of a TFlakShip object.
    - max_fighters_launched
    - rating
    - compensation_rating
    - ending_status */
void
FlakBattle::initShip(TFlakShip& ship)
{
    ship.max_fighters_launched = 2 * ship.bay_count;
    ship.max_fighters_launched = FLAK_MFL_SCALE * ship.bay_count;
    if (ship.max_fighters_launched < FLAK_MINIMUM_MFL)
        ship.max_fighters_launched = FLAK_MINIMUM_MFL;
    if (ship.max_fighters_launched > FLAK_MAXIMUM_MFL)
        ship.max_fighters_launched = FLAK_MAXIMUM_MFL;
    ship.rating              = computeRating(ship);
    ship.compensation_rating = computeStrength(ship);
    ship.ending_status       = 0;
}
