/**
  *  \file game/bum.cc
  *  \brief Beam up Multiple
  *
  *  Beam up multiple is implemented by using two custom transfer
  *  objects, completely unrelated to the normal cargo transfer
  *  partners. The planet half contains 10000 (minus existing command)
  *  of everything; the ship half contains 0 (plus existing command). 
  *  Display is adjusted to actual content using display offsets, so
  *  users can (and should) easily overdraw.
  *
  *  In PCC 1.x, the BUM halves a little closer related, making things
  *  a little more complex but sharing the "init from planet" part. 
  *  PCC 1.x also has a "beam up+down" mode which is a cross of BUM
  *  and a regular cargo unload; I believe this feature to be very
  *  hard to understand, so I left it out here.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/bum.h"
#include "game/ship.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/cmds.h"
#include "game/cspec.h"
#include "game/global.h"
#include "game/config.h"
#include "game/mission.h"
#include "game/fleet.h"

static void
parseBumCommand(int32_t counts[], int sid, int factor)
{
    int shipOwner = getDisplayedTurn().getCurrentUniverse().getShip(sid).getOwner();
    const GCommand* cmd = getDisplayedTurn().getCommands(shipOwner).getCommand(GCommand::phc_Beamup, sid);
    if (!cmd)
        return;

    GCargoSpec sp(cmd->getArg(), true);
    counts[el_Neutronium] += factor * sp.get(el_Neutronium);
    counts[el_Tritanium]  += factor * sp.get(el_Tritanium);
    counts[el_Duranium]   += factor * sp.get(el_Duranium);
    counts[el_Molybdenum] += factor * sp.get(el_Molybdenum);
    counts[el_Colonists]  += factor * sp.get(el_Colonists);
    counts[el_Supplies]   += factor * sp.get(el_Supplies);
    counts[el_Money]      += factor * sp.get(el_Money);
}

/**************************** GShipBumTransfer ***************************/

GShipBumTransfer::GShipBumTransfer(GShip& sh)
    : sh(sh)
{
    init();
}

bool
GShipBumTransfer::canHaveCargo(GCargoType cargo_id)
{
    switch (cargo_id) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
     case el_Money:
     case el_Supplies:
        return true;
     case el_Colonists:
        return config.AllowBeamUpClans();
     default:
        return false;
    }
}

int32_t
GShipBumTransfer::getMaxCargo(GCargoType cargo_id)
{
    const GHull& hull = sh.getHull();
    if (hasFlag(can_Overload)) {
        return 10000;
    } else if (cargo_id == el_Neutronium) {
        return hull.getFuel() - sh.getCargo(el_Neutronium);
    } else if (cargo_id == el_Money) {
        return 10000 - sh.getCargo(el_Money);
    } else {
        int32_t result = sh.getFreeCargo();
        for (int i = 0; i < el_COUNT; ++i)
            if (i != el_Neutronium && i != el_Money && i != cargo_id)
                result -= getCargo(GCargoType(i));
        return result;
    }
}

int32_t
GShipBumTransfer::getDisplayOffset(GCargoType cargo_id)
{
    return display_offsets[cargo_id];
}

string_t
GShipBumTransfer::getName() const
{
    return sh.getName();
}

int32_t
GShipBumTransfer::getCargoOnObject(GCargoType type)
{
    return have[type] - reserved[type];
}

void
GShipBumTransfer::changeCargoOnObject(GCargoType type, int32 amount)
{
    have[type] += amount;
}

void
GShipBumTransfer::changeReservedCargo(GCargoType type, int32 amount)
{
    reserved[type] += amount;
}

void
GShipBumTransfer::commit()
{
    GCargoContainer::commit();

    /* Now we have the cargo amount to transfer in have */
    GCargoSpec cs;
    for (int i = 0; i < el_COUNT; ++i) {
        if (cs.isValidType(GCargoType(i))) {
            cs.set(GCargoType(i), have[i]);
        }
    }

    int sid = sh.getId();
    int shipOwner = getDisplayedTurn().getCurrentUniverse().getShip(sid).getOwner();
    GCommandContainer& cmds = getDisplayedTurn().getCommands(shipOwner);
    if (cs.isEmpty()) {
        cmds.removeCommand(GCommand::phc_Beamup, sid);
        if (sh.getMission() == config.ExtMissionsStartAt() + GMission::pmsn_BeamUpMultiple) {
            setFleetMission(getDisplayedTurn().getCurrentUniverse(), sid, 0, 0, 0);
        }
    } else {
        cmds.setCommand(GCommand::phc_Beamup, sid, cs.toPHostString());
        setFleetMission(getDisplayedTurn().getCurrentUniverse(), sid,
                        config.ExtMissionsStartAt() + GMission::pmsn_BeamUpMultiple, 0, 0);
    }
    sh.markDirty();
}

void
GShipBumTransfer::init()
{
    /* Initialize */
    for (int i = 0; i < el_COUNT; ++i) {
        have[i] = 0;
        display_offsets[i] = 0;
        reserved[i] = 0;
        if (canHaveCargo(GCargoType(i))) {
            display_offsets[i] += sh.getCargo(GCargoType(i));
        }
    }

    /* Add existing command message */
    parseBumCommand(have, sh.getId(), +1);

    setFlags(can_Unload);
}

/*************************** GPlanetBumTransfer **************************/

GPlanetBumTransfer::GPlanetBumTransfer(GPlanet& pl, int sid)
    : pl(pl)
{
    init(sid);
}

bool
GPlanetBumTransfer::canHaveCargo(GCargoType cargo_id)
{
    switch (cargo_id) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
     case el_Money:
     case el_Supplies:
        return true;
     case el_Colonists:
        return config.AllowBeamUpClans();
     default:
        return false;
    }
}

int32_t
GPlanetBumTransfer::getMaxCargo(GCargoType /*cargo_id*/)
{
    return 10000;;
}

int32_t
GPlanetBumTransfer::getDisplayOffset(GCargoType cargo_id)
{
    return display_offsets[cargo_id];
}

string_t
GPlanetBumTransfer::getName() const
{
    return format(_("Beam up from %s"), pl.getName());
}

int32_t
GPlanetBumTransfer::getCargoOnObject(GCargoType type)
{
    return have[type] - reserved[type];
}

void
GPlanetBumTransfer::changeCargoOnObject(GCargoType type, int32 amount)
{
    have[type] += amount;
}

void
GPlanetBumTransfer::changeReservedCargo(GCargoType type, int32 amount)
{
    reserved[type] += amount;
}

void
GPlanetBumTransfer::init(int sid)
{
    /* Initialize */
    for (int i = 0; i < el_COUNT; ++i) {
        have[i] = 10000;
        display_offsets[i] = -10000;
        reserved[i] = 0;
        if (canHaveCargo(GCargoType(i)) && pl.isCargoAmountKnown(GCargoType(i))) {
            display_offsets[i] += pl.getCargo(GCargoType(i));
        }
    }

    /* Add existing command message */
    parseBumCommand(have, sid, -1);

    setFlags(can_GetUnload);
}
