/**
  *  \file game/objtypes.h
  *  \brief Object Type
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_OBJTYPES_H
#define PCC_V2_GAME_OBJTYPES_H

#include "cpluslib/string.h"
#include "cpluslib/sigslot.h"
#include "game/coord.h"

class GUniverse;
class GObject;

class GObjectType {
    GObjectType(const GObjectType&);
    GObjectType& operator=(const GObjectType&);

    int current;
    GUniverse* universe;

 protected:
    /** Initialize. Called by setUniverse(). This may perform any operations the
        object type has to do to adapt to the new universe. Users should call
        setUniverse() instead. Note that this might encounter a null universe. */
    virtual void init() = 0;
    /** Get next object after index. The returned object should exist.
        \param index current object number; might be non-existant or 0. 0 means return first object.
        \returns 0 if no object after that index */
    virtual int getNextIndex(int index) = 0;
    /** Get previous object before index. The returned object should exist.
        \param index current object number; might be non-existant or 0. 0 means return last object.
        \returns 0 if no object before that index */
    virtual int getPreviousIndex(int index) = 0;

    void setEmpty();

 public:
    GObjectType();
    virtual ~GObjectType();

    // Base class interface:

    /** Check whether an object exists. */
    virtual bool isExistant(int index) = 0;
    /** Get object by Id. */
    virtual GObject& getObject(int index) = 0;

    GObject* getCurrentObject();

    int getNextNoWrap(int index, bool mark);
    int getPreviousNoWrap(int index, bool mark);
    int getNextWrap(int index, bool mark);
    int getPreviousWrap(int index, bool mark);
    bool isEmpty();
    bool isUnit();
    int getNumObjects();
    int findNearest(GPoint pt);

    // Operations:

    void setCurrent(int index);
    int getCurrent() const;

    void setUniverse(GUniverse* univ);
    GUniverse* getUniverse() const;

    /** Position changed. Raised when a (possibly) different object is selected.
        Parameter is new index. */
    Signal1<int> sig_pos_changed;
};

class GIndexedObjectType : public GObjectType {
    const int max;

 public:
    GIndexedObjectType(int max);

    // GObjectType:
    void init();
    int getNextIndex(int index);
    int getPreviousIndex(int index);
};

class GAliasObjectType : public GObjectType {
    GObjectType& other;
    SignalConnection conn_pos;

    void onChange();
 public:
    GAliasObjectType(GObjectType& other);

    virtual void init();
    virtual bool isExistant(int index);
    virtual GObject& getObject(int index);
    virtual int getNextIndex(int index);
    virtual int getPreviousIndex(int index);
};

class GObjectTypeObserver {
    GObjectType& type;
    SignalConnection conn_pos;
    SignalConnection conn_obj;

    void onPositionChanged();
 public:
    GObjectTypeObserver(GObjectType& type);
    virtual ~GObjectTypeObserver();
    GObject* getObject();
    GObjectType& getType();

    virtual void onObjectChanged();
};

/** Check for empty type.
    \return true iff this object type contains no unit */
inline bool
GObjectType::isEmpty()
{
    return getCurrent() == 0;
}

/** Get current index. Either 0 or an existing object. */
inline int
GObjectType::getCurrent() const
{
    return current;
}

/** Get current universe. May be null. */
inline GUniverse*
GObjectType::getUniverse() const
{
    return universe;
}

/** Get referred object type. */
inline GObjectType&
GObjectTypeObserver::getType()
{
    return type;
}

#endif
