/**
  *  \file game/ship.h
  *  \brief Ships
  *
  *  This defines all ship-related classes.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_SHIP_H
#define PCC_V2_GAME_SHIP_H

#include <vector>
#include "game/struct.h"
#include "game/obj.h"
#include "game/objtypes.h"
#include "game/spec.h"
#include "game/cargo.h"
#include "cpluslib/maybe.h"
#include "io/io.h"              // mn16_t etc.
#include "game/unitscore.h"
#include "game/playerset.h"

class GUniverse;
class GTurnfile;
class GPlanet;
class GHullFunctionList;

/** Ship. We essentially have three kinds of ship:
    - current ship, i.e. one seen this turn, and possibly played
    - guessed ship, i.e. not seen this turn but we guess it's there
    - history ship, i.e. not seen this turn */
class GShip : public GMapObject {
 private:
    GShip();
    GShip(const GShip&);
    GShip& operator=(const GShip&);
 public:
    /// Transporters.
    enum Transporter {
        t_Unload,
        t_Transfer
    };
    enum Kind {
        NoShip,             /**< We do not know anything about this ship. */
        CurrentShip,        /**< This is a current ship. It is visible, we have a SHIP.DAT entry for it. */
        CurrentTarget,      /**< This is a current ship. It is visible, we have a TARGET.DAT entry for it. */
        CurrentUnknown,     /**< This is a current ship. It is visible, but we don't have any data but its mass (non-visual contact). */
        GuessedShip,        /**< This is a guessed ship. We have no current data for it, just history, and we display that on the map. */
        HistoryShip         /**< This is an old ship. We have history data for it, and it's not visible this turn. */
    };

 private:
    struct MapInfo {
        int id;             /**< Ship Id. Constant. */
        GPoint pos;         /**< Position. Known for CurrentShip, CurrentTarget, CurrentUnknown, GuessedShip. Otherwise 0. */
        int owner;          /**< Owner. Known for CurrentShip, CurrentTarget, CurrentUnknown, GuessedShip. Otherwise 0. */
        int scanned_mass;   /**< Scanned mass. Known for CurrentTarget and CurrentUnknown. For CurrentShip, taken from ship_data. */
        int orbit;          /**< Planet we're orbiting. 0 if none. Known if position known. */
        int rc_flag;        /**< Remote control. 0 if not RCd, -1 if forbidden, otherwise real owner. */
        Kind kind : 8;      /**< Kind of ship. */

        MapInfo(int id);
    };

    MapInfo map_info;

    // TODO: replace ship_data and target_data by opaque structures which also
    // contain explicit history fields, similar to planets.
    struct ShipInfo;

    /** Ship data. For CurrentShip, nonnull and contains current data.
        For all other types but NoShip, nonnull and contains history data. */
    ShipInfo* ship_info;

    /** Ship target data. Valid for CurrentTarget and possibly
        CurrentShip, contains the target data. This member is (almost)
        never accessed outside initialisation. The only field we fetch
        from here is TShipTarget::heading, all other fields are copied
        into ship_data upon initialisation. */
    TShipTarget* target_data;

    /** Source flags. Bitfield of who's .dat file the information is
        from. Used for saving. */
    GPlayerSet src_flags;

    /** Target flags. Bitfield of who's TARGET file the information is
        from. Used for rule checks. */
    GPlayerSet target_flags;

    /** Special functions. */
    std::vector<int16_t> special_functions;

    /** Cargo arbiter. */
    GCargoArbiter cargo;

    /** Transporter in-use flags. */
    bool transporter_in_use[2];

    /** Fleet */
    int fleet_nr;
    string_t fleet_name;

    ShipInfo& createShip();
    const TShip& getDisplayedShip() const;
    void adjustShipTrack(int turn);
 public:
    GUnitScores unit_scores;

    GShip(int id);
    ~GShip();

    /* Loading and Saving */
    void addShipData(const TShip& dat, GPlayerSet srcf);
    void addTargetData(const TShipTarget& target, GPlayerSet srcf);
    void addShipXYData(TShipXY axy);
    void addRCEntry(int flag);
    void addShipHistoryData(const TDbShip& ship);
    void addShipTrackEntry(const TDbShipTrackEntry& entry, int turn);
    void addShipFunction(int function);

    void getShipData(TShip& the_data) const;
    int  getShipTrackHeader(TDbShipTrack& header) const;
    void getShipTrackEntry(TDbShipTrackEntry& entry, int slot) const;
    void getShipHistoryData(TDbShip& the_data) const;

    /* Turn access: these are in game/trn-cmd.cc */
    void applyTurnCommand(GTurnfile& trn, int index);
    void makeTurnCommands(GTurnfile& trn, const GShip& dis) const;

    /* GMapObject interface */
    string_t getName() const;
    string_t getFullName() const;
    int      getId() const;
    int      getOwner() const;
    bool     isExistant() const;
    GPoint   getPos() const;

    /* Checking */
    void internalCheck();
    void combinedCheck1(GUniverse& univ);

    /* Accessors: status */
    bool hasShipData() const;
    bool hasAnyData() const;
    bool isReliablyVisible(int for_player = 0) const;
    Kind getKind() const;
    GPlayerSet getSrcFlags() const;
    GPlayerSet getTargetFlags() const;

    int16_t getHistoryTimestamp(TDbShipTimestamp which_one) const;

    /* Accessors: type/location */
    int getMass() const;
    int getHullId() const;
    GHull& getHull() const;
    int getOrbit() const;
    int getRealOwner() const;
    int getRemoteControlFlag() const;

    /* Accessors: specials */
    bool canDoSpecial(int basic_function) const;
    void enumShipSpecificSpecials(GHullFunctionList& list) const;
    bool hasAnyShipSpecificSpecials() const;

    /* Accessors: course */
    GPoint getWaypoint() const;
    void   setWaypoint(GPoint pt);
    int getWaypointDX() const;
    int getWaypointDY() const;
    mp16_t getHeading() const;

    mp16_t getWarp() const;
    void   setWarp(int warp);
    bool isHyperdriving() const;

    /* Accessors: Equipment */
    mp16_t getEngine() const;
    void   setEngine(mp16_t engine);
    mp16_t getBeamType() const;
    void   setBeamType(mp16_t type);
    mp16_t getBeamCount() const;
    void   setBeamCount(mp16_t count);
    mp16_t getBayCount() const;
    void   setBayCount(mp16_t count);
    mp16_t getTorpType() const;
    void   setTorpType(mp16_t type);
    mp16_t getTorpLauncherCount() const;
    void   setTorpLauncherCount(mp16_t count);

    /* Accessors: Mission */
    void   setName(const string_t& str);
    mp16_t getMission() const;
    void   setMission(int m, int i, int t);
    mp16_t getTowId() const;
    mp16_t getIntId() const;
    mp16_t getPE() const;
    void   setPE(int pe);
    mp16_t getDamage() const;
    void   setDamage(mp16_t damage);
    mp16_t getCrew() const;
    void   setCrew(mp16_t crew);
    string_t getFCode() const;
    void   setFCode(const string_t& str);
    bool   isFCodeKnown() const;

    /* Accessors: Cargo */
    int getAmmo() const;
    int getColonists() const;
    int getSupplies() const;
    int getMoney() const;
    int getCargoRaw(GCargoType type) const;
    int getCargo(GCargoType type) const;
    void changeCargoRaw(GCargoType type, int amount);
    int getFreeCargo() const;

    GCargoArbiter& getCargoArbiter();

    /* Accessors: Transporters */
    TShipTransfer& getTransporter(Transporter which_one) throw();
    const TShipTransfer& getTransporter(Transporter which_one) const throw();
    bool isTransporterActive(Transporter which_one) const throw();
    bool isTransporterCancellable(Transporter which_one) const throw();
    void cancelTranporter(Transporter which_one) throw();
    bool isTransporterInUse(Transporter which_one) const throw();
    void setTransporterInUse(Transporter which_one, bool in_use) throw();

    bool isCloningAt(const GPlanet& planet) const;

    /* Accessors: History */
    const TDbShipTrackEntry& getHistoryEntry(int turn) const;
    int getHistoryNewestTurn() const;
    int getHistoryOldestTurn() const;

    /* Accessors: Fleet */
    void setFleetNumber(int fno);
    int getFleetNumber() const;
    void setFleetName(string_t name);
    const string_t& getFleetName() const;
    bool isFleetLeader() const;
    bool isFleetMember() const;
};

class GShipType : public GIndexedObjectType {
 public:
    GShipType();

    bool isExistant(int index);
    GShip& getObject(int index);
};

class GAnyShipType : public GIndexedObjectType {
 public:
    GAnyShipType();

    bool isExistant(int index);
    GShip& getObject(int index);
};

class GHistoryShipType : public GIndexedObjectType {
 public:
    GHistoryShipType();

    bool isExistant(int index);
    GShip& getObject(int index);
};

/** Ship Cargo Transfer. Describes the ship's main cargo hold for
    cargo transfer and transaction purposes. */
class GShipTransfer : public GCargoContainer {
    GShip& ship;
 protected:
    int32  getCargoOnObject(GCargoType type);
    void   changeCargoOnObject(GCargoType type, int32 amount);
    void   changeReservedCargo(GCargoType type, int32 amount);
 public:
    GShipTransfer(GShip& ship);
    ~GShipTransfer();

    bool     canHaveCargo(GCargoType cargo_id);
    int32    getMaxCargo(GCargoType cargo_id);
    string_t getName() const;
    int32    getDisplayOffset(GCargoType type);

    GShip&   getShip();
};

/** Ship Cargo Transporter. Describes the ship/ship or ship/planet
    (resp. ship/space) transporter. For each ship and each
    transporter, only one object of this type can exist at any given
    time. */
class GShipTransporter : public GCargoContainer {
    GShip&             ship;     //!< The ship whose transporter we're using.
    TShipTransfer&     transfer; //!< The transporter we're using.
    GObject*           other;    //!< The other object (null if jettison).
    GShip::Transporter which_one;//!< Transporter identifier.
    TShipTransfer      reserve;  //!< Reservations, to handle GCargoContainer protocol.
 protected:
    int32  getCargoOnObject(GCargoType type);
    void   changeCargoOnObject(GCargoType type, int32 amount);
    void   changeReservedCargo(GCargoType type, int32 amount);
 public:
    GShipTransporter(GShip& ship, GShip::Transporter which_one, GObject& other);
    GShipTransporter(GShip& ship);
    ~GShipTransporter();

    bool     canHaveCargo(GCargoType cargo_id);
    int32    getMaxCargo(GCargoType cargo_id);
    string_t getName() const;
    int32    getDisplayOffset(GCargoType type);
};

/** Ship Turn Predictor. This object manages prediction of a ship's
    future, especially movement fuel and ETA computation.

    To compute prediction for a single ship, use the two-argument
    constructor. To predict a tower/towee pair, construct a secondary
    object for the towee, and pass it in to the tower's predictor
    using the three-argument constructor. The tower's computeTurn()
    method will then also compute the towee's prediction. */
class GShipTurnPredictor {
    int                 id;
    TShip               ship;
    bool                valid;
    GShipTurnPredictor* towee_override;
    const GUniverse&    univ;
    int32_t             movement_fuel_used;
    int32_t             cloak_fuel_used;
    int                 turns_computed;

 public:
    GShipTurnPredictor(const GUniverse& turn, int id);
    GShipTurnPredictor(const GUniverse& turn, int id, GShipTurnPredictor& towee_override);

    /** Get total fuel used for movement. */
    int32 getMovementFuelUsed() const
        { return movement_fuel_used; }
    int32_t getCloakFuelUsed() const
        { return cloak_fuel_used; }

    /** Get number of turns computed. */
    int getTurnCount() const
        { return turns_computed; }

    void  computeTurn();
    void  computeMovement();

    void  setPosition(GPoint pt);
    void  setWaypoint(GPoint pt);
    void  setSpeed(int warp);
    void  setEngine(int engine);
    void  setMission(int m, int i, int t);
    void  setFCode(string_t s);
    void  setFuel(int fuel);

    bool  isAtWaypoint() const;
    GPoint getPosition() const;
    int   getFuel() const;
    int   getSpeed() const;
    int   getRealOwner() const;
    string_t getFCode() const;
    const GUniverse& getUniverse() const;

    static const int MOVEMENT_TIME_LIMIT = 30;
};

/****************************** Mine Laying ******************************/

struct GMineMissionData {
    int   mine_id;            /**< Id of mine field to manipulate. 0 to accept any. */
    int   owner;              /**< Required owner. */
    bool  type;               /**< Type of minefield. */
    int   num_torps;          /**< Number of torpedoes. */
    int32 num_units;          /**< Number of mine units. */
};

bool parseMineLayingMission(const GShip& sh, const GUniverse& univ, GMineMissionData& md);
bool parseMineScoopMission(const GShip& sh, const GUniverse& univ, GMineMissionData& md);

/******************************* Chunneling ******************************/

struct GChunnelData {
    int   target;
    int   failure;
    int   kind;

    /* Failure modes. Note that the sequence of these values is used
       in parseChunnelMission. */
    enum {
        chf_MateDamaged = 1,
        chf_MateMoving  = 2,
        chf_MateTowed   = 4,
        chf_MateFuel    = 8,
        chf_MateAny     = 15,

        chf_Damaged     = 16,
        chf_Moving      = 32,
        chf_Towed       = 64,
        chf_Fuel        = 128,
        chf_SelfAny     = 240,

        chf_Distance    = 256
    };

    /* Kinds */
    enum {
        chk_Self   = 1,
        chk_Others = 2
    };
};

bool parseChunnelMission(const GShip& sh, const GUniverse& univ, GChunnelData& cd);

/**************************** Inline Functions ***************************/

/** Check whether we have full, playable data. */
inline bool
GShip::hasShipData() const
{
    return map_info.kind == CurrentShip;
}

/** Check whether we have any data about this ship. */
inline bool
GShip::hasAnyData() const
{
    return ship_info != 0;
}

/** Get kind of this ship. */
inline GShip::Kind
GShip::getKind() const
{
    return map_info.kind;
}

/** Get ship source flags. This is the set of players whose SHIP
    file contains a copy of this ship (usually a unit set). */
inline GPlayerSet
GShip::getSrcFlags() const
{
    return src_flags;
}

/** Get target source flags. This is the set of players whose TARGET
    file contains a copy of this ship, and who therefore reliably know
    about this ship's existance. Should be disjoint from getSrcFlags().

    FIXME: what about nonvisual contacts (SHIPXY)? */
inline GPlayerSet
GShip::getTargetFlags() const
{
    return target_flags;
}

/** Get planet we're orbiting.
    \return planet id, 0 if none. */
inline int
GShip::getOrbit() const
{
    return map_info.orbit;
}

/** Get ship's remote control flag. */
inline int
GShip::getRemoteControlFlag() const
{
    return map_info.rc_flag;
}

/** Get cargo arbiter for this ship. This manages reservations for the
    ship's main cargo hold. The transporters (ship/ship, ship/planet)
    do not have a cargo arbiter. They are managed by GShipTransporter,
    of which only one instance can be active per transporter and ship
    at any given time (see GShip::isTransporterInUse). */
inline GCargoArbiter&
GShip::getCargoArbiter()
{
    return cargo;
}

inline void
GShip::setFleetNumber(int fno)
{
    this->fleet_nr = fno;
}

inline int
GShip::getFleetNumber() const
{
    return fleet_nr;
}

inline void
GShip::setFleetName(string_t name)
{
    this->fleet_name = name;
}

inline const string_t&
GShip::getFleetName() const
{
    return fleet_name;
}

inline bool
GShip::isFleetLeader() const
{
    return fleet_nr == map_info.id;
}

inline bool
GShip::isFleetMember() const
{
    return fleet_nr != 0 && fleet_nr != map_info.id;
}



inline GShip&
GShipTransfer::getShip()
{
    return ship;
}


void setOptimumWarp(GUniverse& univ, int sid);

#endif
