/**
  *  \file spec.h
  *
  *  Specification files. This file contains the classes which
  *  correspond to static game information (hullspec etc.)
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_SPEC_H
#define PCC_V2_SPEC_H

#include <string>
#include <vector>
#include "game/struct.h"
#include "cpluslib/assert.h"
#include "game/hullfunc.h"
#include "game/exp.h"
#include "game/cost.h"
#include "game/playerset.h"
#include "cpluslib/ptrvector.h"
#include "util/modinit.h"

/* Starship Component */
class GComponent {
    int id;
 public:
    /** Constructor. */
    GComponent(int id)
        : id(id)
        { }
    virtual ~GComponent();

    /** Get mass of this component, in kilotons. */
    virtual int getMass() const = 0;
    /** Get tech level required to build this component. */
    virtual int getTech() const = 0;
    /** Get cost of one of this components. */
    virtual GCost getCost() const = 0;
    /** Get name of this component. This is the full name. */
    virtual string_t getName() const = 0;
    /** Get short name of this component. */
    virtual string_t getShortName() const = 0;

    /** Get Id number of this component. */
    int getId() const
        { return id; }
};

/* Weapon */
class GWeapon : public GComponent {
 public:
    /** Constructor. */
    GWeapon(int id)
        : GComponent(id) { }
    /** Get kill (anti-life) power. Should not be 'AC-mangled'. */
    virtual int getKill() const = 0;
    /** Get damage (anti-structuer) power. */
    virtual int getDamage() const = 0;
};

/* Beam */
class GBeam : public GWeapon {
    TBeam data;
 public:
    GBeam(int aid, const TBeam& adata);

    string_t getName() const;
    string_t getShortName() const;

    int getKill() const
        { return data.kill; }
    int getDamage() const
        { return data.damage; }
    int getMass() const
        { return data.mass; }
    int getTech() const
        { return data.tech; }
    GCost getCost() const
        { return data.cost; }

    /** Get damage value. This is a fast access path bypassing vtbls. */
    int f_getDamage() const
        { return data.damage; }
    /** Get kill value. This is a fast access path bypassing vtbls. */
    int f_getKill() const
        { return data.kill; }
};

/* Torpedo */
class GTorpedo : public GWeapon {
    friend class GTorpLauncher;
    TTorpedo data;

    GTorpedo(int id, const TTorpedo& data);
 public:
    string_t getName() const;
    string_t getShortName() const;

    int getKill() const
        { return data.kill; }
    int getDamage() const
        { return data.damage; }
    int getMass() const
        { return 1; }
    int getTech() const
        { return data.tech; }
    GCost getCost() const;
};

/* Torpedo Launcher */
class GTorpLauncher : public GWeapon {
    GTorpedo torp;
 public:
    GTorpLauncher(int id, const TTorpedo& data);

    string_t getName() const;
    string_t getShortName() const;

    int getKill() const
        { return torp.data.kill; }
    int getDamage() const
        { return torp.data.damage; }
    int getMass() const
        { return torp.data.laun_mass; }
    int getTech() const
        { return torp.data.tech; }
    GCost getCost() const
        { return torp.data.laun_cost; }

    /** Get cost of one torpedo in mc. */
    int getTorpCost() const
        { return torp.data.torp_cost; }

    /** Get torpedo. */
    const GTorpedo& getTorpedo()
        { return torp; }

    int f_getDamage() const
        { return torp.data.damage; }
    int f_getKill() const
        { return torp.data.kill; }
};

/* Fighter */
class GFighter : public GWeapon {
 public:
    GFighter(int id)
        : GWeapon(id) { }
    string_t getName() const;
    string_t getShortName() const;
    int getKill() const;
    int getDamage() const;
    int getMass() const
        { return 1; }
    int getTech() const
        { return 1; }
    GCost getCost() const;
};

/* Engine */
class GEngine : public GComponent {
    TEngine data;
    int max_eff_warp;
 public:
    GEngine(int aid, const TEngine& adata);

    int getMass() const
        { return 0; }
    int getTech() const
        { return data.tech; }
    GCost getCost() const
        { return data.cost; };
    string_t getName() const;
    string_t getShortName() const;

    int getFuelFactor(int warp) const;

    int getMaxEfficientWarp() const;

    void setMaxEfficientWarp(int warp);
};

/* Starship Hull */
class GHull : public GComponent {
    THull data;

    struct SpecialAssignment {
        GHullFunctionData::dev_t device;
        GPlayerSet added;
        GPlayerSet removed;

        SpecialAssignment(GHullFunctionData::dev_t device, GPlayerSet added, GPlayerSet removed)
            : device(device), added(added), removed(removed)
        { }
    };

    typedef std::vector<SpecialAssignment> special_t;

    special_t specials[2];

    SpecialAssignment* getSpecialFunctionPtr(GHullFunctionData::dev_t, bool assign_to_hull);

    void enumerateSet(const special_t& stuff, GHullFunctionList& asgn, GPlayerSet player_restriction,
                      GExpLevelSet level_restriction, GHullFunction::Kind kind) const;

    GPlayerSet checkSet(const special_t& stuff, int basic_function,
                        GExpLevelSet levels, bool use_defaults) const;
 public:
    GHull(int aid, const THull& adata);
    ~GHull();

    int getMass() const
        { return data.mass; }
    int getTech() const
        { return data.tech; }
    GCost getCost() const;
    string_t getName() const;
    string_t getShortName() const;

    /** Get picture number. \todo specify some properties of this value. */
    int getPicture() const
        { return data.picture; }
    /** Get size of fuel tank. */
    int getFuel() const
        { return data.fuel; }
    /** Get size of normal crew on this ship. */
    int getCrew() const
        { return data.crew; }
    /** Get number of engines on this ship. */
    int getNumEngines() const
        { return data.engines; }
    /** Get size of cargo room on this ship. */
    int getCargo() const
        { return data.cargo; }
    /** Get number of bays on this ship. */
    int getNumBays() const
        { return data.bays; }
    /** Get maximum number of torpedo launchers on this ship. */
    int getMaxLaunchers() const
        { return data.launchers; }
    /** Get maximum number of beam weapons on this ship. */
    int getMaxBeams() const
        { return data.beams; }

    void changeSpecialFunction(GHullFunctionData::dev_t id, GPlayerSet add, GPlayerSet remove, bool assign_to_hull);

    void enumerateHullFunctions(GHullFunctionList& asgn,
                                GPlayerSet player_restriction,
                                GExpLevelSet level_restriction,
                                bool include_new_ship,
                                bool include_racial_abilities) const;

    void clearSpecialFunctions();
    bool canDoSpecial(int basic_function, int player, bool hull, GExpLevelSet levels) const;
    GPlayerSet getPlayersThatCan(int basic_function, bool hull, GExpLevelSet levels) const;

    static special_t racial_abilities;

    static void clearRacialAbilities();
    static void findRacialAbilities();
};

extern ptr_vector<GHull> hull_spec;
extern TTruehull truehull_spec;

extern const ModuleInitializer mod_spec[];

GBeam&         getBeam(int bid);
GTorpLauncher& getTorp(int tid);
GEngine&       getEngine(int eid);
GHull&         getHull(int hid);
int            getTruehull(int pid, int slot);
int            getTruehullSlot(int pid, int hid);

/** Get fighter for a player.
    \todo maybe cache the objects somehow? */
inline GFighter
getFighterForPlayer(int pid)
{
    return GFighter(pid);
}

/** Get number of hulls. */
inline int
getNumHulls()
{
    return hull_spec.size();
}

#endif
