/**
  *  \file int/autotask.cc
  *  \brief Interpreter: Editing of Auto Tasks
  *
  *  An auto task is presented to the user as a list of commands
  *  (strings). Internally, those commands are compiled into a
  *  sequence of interpreter instructions:
  *  - regular commands produce
  *       pushlit 'the command'
  *       pushlit CC$AUTOEXEC
  *       callind 1
  *  - the 'restart' command produces
  *       pushlit CC$AUTORECHECK
  *       callind 0
  *       j #0
  *
  *  This module provides a class IntAutoTaskEditor to convert to and
  *  from these formats.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "int/autotask.h"
#include "int/bytecode.h"
#include "int/opcode.h"
#include "int/intglobal.h"
#include "int/exec.h"
#include "int/lex.h"
#include "int/keywords.h"

/******************************* Predicates ******************************/

/* For decompilation: */

/** Check for 'pushlit' instruction. */
static bool
isPushLiteral(const IntBytecodeObject& bco, IntBytecodeObject::pc_t pc)
{
    const IntOpcode& opc = bco(pc);
    return opc.major == IntOpcode::maPush
        && opc.minor == IntOpcode::sLiteral;
}

/** Check for instruction that pushes a global variable. Accepted encodings are:
    - pushvar NAME
    - pushgvar NAME
    - pushglob #NAME */
static bool
isPushGlobal(const IntBytecodeObject& bco, IntBytecodeObject::pc_t pc, const char* name)
{
    const IntOpcode& opc = bco(pc);
    return opc.major == IntOpcode::maPush
        && (((opc.minor == IntOpcode::sNamedVariable
              || opc.minor == IntOpcode::sNamedShared)
             && bco.getName(opc.arg) == name)
            || (opc.minor == IntOpcode::sShared
                && global_property_names.lookupAddress(opc.arg) == name));
}

/** Check for 'callind NARGS' instruction. */
static bool
isCall(const IntBytecodeObject& bco, IntBytecodeObject::pc_t pc, uint16_t nargs)
{
    const IntOpcode& opc = bco(pc);
    return opc.major == IntOpcode::maIndirect
        && (opc.minor & ~(IntOpcode::miIMRefuseFunctions | IntOpcode::miIMRefuseProcedures)) == IntOpcode::miIMCall
        && opc.arg == nargs;
}

/** Check for 'j #0' instruction. */
static bool
isJump0(const IntBytecodeObject& bco, IntBytecodeObject::pc_t pc)
{
    const IntOpcode& opc = bco(pc);
    return opc.major == IntOpcode::maJump
        && opc.minor == IntOpcode::jAlways
        && opc.arg == 0;
}

/*************************** IntAutoTaskEditor ***************************/

IntAutoTaskEditor::IntAutoTaskEditor()
    : pc(0), pc_rel(0)
{ }

IntAutoTaskEditor::~IntAutoTaskEditor()
{ }

/** Load from process. Converts the process into an editable string
    list, stored in this object.
    \retval true Conversion succeeded
    \retval false Conversion failed; this probably is not an auto task */
bool
IntAutoTaskEditor::loadFromProcess(IntExecutionContext& exc)
{
    /* No need to validate exception frames or contexts; those can
       only be produced when the code contains appropriate
       instructions, and if it does, it fails our syntax check. */
    code.clear();
    pc = 0;
    pc_rel = 0;

    /* Must have at least one frame. Zero frames means an empty
       process, which is a valid auto task. */
    if (exc.getNumActiveFrames() < 1) {
        return true;
    }
    IntExecutionFrame* frame = exc.getOutermostFrame();
    IntBCORef bco = frame->bco;

    /* Parse the code. Note that I try to avoid the assumption that
       'Restart' and regular commands are compiled to instructions
       of the same length, so this could easier be changed later. */
    IntBytecodeObject::pc_t raw_pc = 0;
    while (IntBytecodeObject::pc_t remaining = bco->getNumInstructions() - raw_pc) {
        /* Check for regular command */
        if (remaining >= 3
            && isPushLiteral(*bco, raw_pc)
            && isPushGlobal(*bco, raw_pc+1, "CC$AUTOEXEC")
            && isCall(*bco, raw_pc+2, 1))
        {
            IntStringValue* sv = dynamic_cast<IntStringValue*>(bco->getLiteral((*bco)(raw_pc).arg));
            if (sv != 0) {
                checkSetPC(exc, raw_pc, 3);
                code.push_back(sv->getValue());
                raw_pc += 3;
                continue;
            }
        }

        /* Check for Restart command */
        if (remaining >= 3
            && isPushGlobal(*bco, raw_pc, "CC$AUTORECHECK")
            && isCall(*bco, raw_pc+1, 0)
            && isJump0(*bco, raw_pc+2))
        {
            checkSetPC(exc, raw_pc, 3);
            code.push_back("Restart");
            raw_pc += 3;
            continue;
        }

        /* When we're here, the instruction cannot be interpreted. Fail. */
        clear();
        return false;
    }

    return true;
}

/** Save to process. Compiles this string list back into a process.
    The process will usually be an already existing one, which will be
    modified accordingly. */
void
IntAutoTaskEditor::saveToProcess(IntExecutionContext& exc)
{
    /* Generate new BCO */
    IntBCORef bco = new IntBytecodeObject();
    bco->setIsProcedure(true);
    bco->setName(exc.getName());
    IntBytecodeObject::pc_t new_pc = 0;
    for (uint32_t i = 0; i < code.size(); ++i) {
        /* Is this the new program counter? */
        if (i == pc) {
            new_pc = bco->getNumInstructions() + pc_rel;
        }
        /* Generate code */
        if (isRestartCommand(code[i])) {
            /* Encode restart operation */
            bco->addInstruction(IntOpcode::maPush, IntOpcode::sNamedShared, bco->addName("CC$AUTORECHECK"));
            bco->addInstruction(IntOpcode::maIndirect, IntOpcode::miIMCall, 0);
            bco->addInstruction(IntOpcode::maJump, IntOpcode::jAlways, 0);
        } else {
            /* Encode normal operation */
            IntStringValue sv(code[i]);
            bco->addPushLiteral(&sv);
            bco->addInstruction(IntOpcode::maPush, IntOpcode::sNamedShared, bco->addName("CC$AUTOEXEC"));
            bco->addInstruction(IntOpcode::maIndirect, IntOpcode::miIMCall, 1);
        }
    }

    /* Check active frames */
    if (exc.getNumActiveFrames() < 1) {
        /* No frame at all. This means the process was newly created. */
        exc.pushNewFrame(new IntExecutionFrame(bco));
    } else if (exc.getNumActiveFrames() > 1 && pc_rel == 0) {
        /* We're inside a call, but the new PC is outside. Drop all frames. */
        while (exc.getNumActiveFrames() > 1) {
            exc.popFrame();
        }
        /* popFrame does not pop the value stack. Since we're at the beginning
           of an instruction sequence, the stack ought to be empty. */
        while (exc.getStackSize() > 0) {
            exc.dropValue();
        }
    } else {
        /* Don't change the frame sequence */
    }

    /* Fix up outermost frame */
    IntExecutionFrame* frame = exc.getOutermostFrame();
    frame->pc = new_pc;
    frame->bco = bco;
}

/** Clear this editor. */
void
IntAutoTaskEditor::clear()
{
    code.clear();
    pc = 0;
    pc_rel = 0;
}

/** Insert/replace a line.
    \param pos Position to modify [0,getNumInstructions()]
    \param nold Insert (nold=0) or replace (nold=1)
    \param line Line to insert/replace */
void
IntAutoTaskEditor::replace(uint32_t pos, uint32_t nold, const string_t& line)
{
    replace(pos, nold, &line, 1);
}

/** Generic update. Replace \c nold lines starting at \c pos by \c
    nnew lines from \c line. This can be used for insertion (nold=0),
    deletion (nnew=0), or replacement in any combination.

    \param pos Position to modify [0,getNumInstructions()]
    \param nold Number of instructions to erase
    \param line Instructions to insert (can be 0 if \c nnew=0)
    \param nnew Number of instructions to insert  */
void
IntAutoTaskEditor::replace(uint32_t pos, uint32_t nold, const string_t line[], uint32_t nnew)
{
    /* Validate */
    pos = std::min(pos, code.size());
    nold = std::min(nold, code.size() - pos);

    /* Update vector size */
    if (nnew > nold) {
        code.insert(code.begin() + pos, nnew - nold, string_t());
    }
    if (nnew < nold) {
        code.erase(code.begin() + pos, code.begin() + pos + nold - nnew);
    }

    /* Update vector content */
    for (uint32_t i = 0; i < nnew; ++i) {
        code[pos + i] = line[i];
    }

    /* Handle PC */
    if (pc >= pos && pc < pos+nold) {
        /* PC is within modified area */
        pc = pos;
        pc_rel = 0;
    } else if (pc >= pos+nold) {
        /* PC is after modified area */
        if (pos == 0 && code.size() <= nnew) {
            /* Special case: we have inserted after the end */
            pc = 0;
            pc_rel = 0;
        } else {
            pc -= nold;
            pc += nnew;
        }
    }
}

/** Set program counter. This will set the PC to outside the specified
    instruction. */
void
IntAutoTaskEditor::setPC(uint32_t newPC)
{
    if (newPC < code.size()) {
        pc = newPC;
        pc_rel = 0;
    }
}

/** Check whether a command is allowed in an auto task. Refuses
    commands that are syntactically invalid, and commands which are
    obviously not procedure calls.

    We have to refuse multi-line commands because they obviously will
    not work when wrapped into CC$AUTOEXEC calls line-by-line. To
    avoid the need to reliably distinguish one-line and multi-line, we
    refuse structural commands completely. This is the same
    restriction as in PCC 1.x, although for a different reason (in
    1.x, structural commands affect the runtime context stack in a way
    the editor cannot handle). We refuse a few commands more than PCC
    1.x, but the additional commands would not have worked in 1.x as
    well.

    \param cmd Command to check */
bool
IntAutoTaskEditor::isValidCommand(const string_t& cmd)
{
    try {
        IntTokenizer tok(cmd);

        // Allow empty commands. Why not.
        if (tok.getCurrentToken() == tok.tEnd) {
            return true;
        }

        // Commands must start with an identifier.
        // (PCC 1.x also accepts anything starting with a non-identifier.)
        if (tok.getCurrentToken() != tok.tIdentifier) {
            return false;
        }

        // Reject commands that could possibly be multi-line.
        // This is a superset of the PCC 1.x restriction.
        switch (lookupKeyword(tok.getCurrentString())) {
         case kwBreak:          // not in 1.x
         case kwCase:
         case kwContinue:       // not in 1.x
         case kwDo:
         case kwElse:
         case kwEndFunction:    // not in 1.x
         case kwEndIf:
         case kwEndSelect:
         case kwEndSub:
         case kwEndTry:
         case kwEndWith:
         case kwFor:
         case kwForEach:
         case kwFunction:       // not in 1.x
         case kwIf:
         case kwLoop:
         case kwNext:
         case kwReturn:         // not in 1.x
         case kwSelect:
         case kwSub:
         case kwTry:
         case kwUntil:          // not in 1.x
         case kwWhile:          // not in 1.x
         case kwWith:
            return false;
         default:
            break;
        }

        // While we have a tokenizer handy, check that the whole line tokenizes well.
        while (tok.getCurrentToken() != tok.tEnd) {
            if (tok.getCurrentToken() == tok.tInvalid) {
                return false;
            }
            tok.readNext();
        }
        return true;
    }
    catch (IntError&) {
        return false;
    }
}

/** Check for 'Restart' command. */
bool
IntAutoTaskEditor::isRestartCommand(const string_t& cmd)
{
    try {
        IntTokenizer tok(cmd);
        return tok.getCurrentToken() == tok.tIdentifier
            && tok.getCurrentString() == "RESTART"
            && tok.readNext() == tok.tEnd;
    }
    catch (...) {
        return false;
    }
}

/** Check for blank command. */
bool
IntAutoTaskEditor::isBlankCommand(const string_t& cmd)
{
    try {
        IntTokenizer tok(cmd);
        return tok.getCurrentToken() == tok.tEnd;
    }
    catch (...) {
        return false;
    }
}

/** Decompiler: Check and set program counter from parsed process.
    This will place the user PC at the end of /code/ when the real PC
    is within the range described by raw_pc/length. */
void
IntAutoTaskEditor::checkSetPC(IntExecutionContext& ctx, uint32_t raw_pc, uint32_t length)
{
    IntBytecodeObject::pc_t real_pc = ctx.getOutermostFrame()->pc;
    bool found = false;
    if (ctx.getNumActiveFrames() == 1) {
        /* One frame. */
        if (real_pc >= raw_pc && real_pc < raw_pc + length) {
            /* real PC is within interval */
            found = true;
        }
    } else {
        /* More than one frame. */
        if (real_pc > raw_pc && real_pc <= raw_pc + length) {
            /* real PC is one-after an instruction within the interval */
            found = true;
        }
    }

    if (found) {
        pc = code.size();
        pc_rel = real_pc - raw_pc;
    }
}
