/**
  *  \file int/builtin.cc
  *  \brief Interpreter: Code Generation for Builtin Functions
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include "int/builtin.h"
#include "cpluslib/misc.h"
#include "int/expr.h"
#include "cpluslib/types.h"
#include "int/unary.h"
#include "int/binary.h"
#include "int/interror.h"

namespace {
    enum {
        FC_Generic,
        FC_Ship,
        FC_Planet
    };

    /** "If" function. This binary or ternary function does not
        evaluate all its arguments. */
    class IntIfFunctionCallNode : public IntFunctionCallNode {
     public:
        void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    };

    /** Function that "folds" any number of arguments into a single value using
        a binary operation. Used for "Min" and "Max" functions as well as bitwise
        operations.

        This has special cases for Min/Max (to handle the CaseBlind flag), and for
        BitOr/BitAnd/BitXor (to enforce type safety when used with just one arg),
        but could otherwise be used to fold anything. */
    class IntFoldFunctionCallNode : public IntFunctionCallNode {
        uint8_t minor;
     public:
        IntFoldFunctionCallNode(uint8_t minor)
            : minor(minor)
            { }
        void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    };

    /** Regular builtin function. Generates code for unary or binary functions.
        Caller must make sure that it receives an appropriate number of parameters. */
    class IntBuiltinFunctionNode : public IntFunctionCallNode {
        uint8_t minor;
     public:
        IntBuiltinFunctionNode(uint8_t minor)
            : minor(minor)
            { }
        void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    };

    /** Regular builtin binary function that can distinguish case.
        For case-blind operation, generates opcode minor+1. */
    class IntCaseFunctionNode : public IntFunctionCallNode {
        uint8_t minor;
     public:
        IntCaseFunctionNode(uint8_t minor)
            : minor(minor)
            { }
        void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    };

    /** "Find" function. Iterates over its first arg (an array),
        looking for an object where the second arg (an expression) yields
        true, returning third arg. */
    class IntFindFunctionCallNode : public IntFunctionCallNode {
        uint8_t which;
     public:
        IntFindFunctionCallNode(uint8_t which)
            : which(which)
            { }
        void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    };

    /** "Count" function. Iterates over its first arg (an array), and
        counts objects where the the second arg (an expression) yields
        true. */
    class IntCountFunctionCallNode : public IntFunctionCallNode {
        uint8_t which;
     public:
        IntCountFunctionCallNode(uint8_t which)
            : which(which)
            { }
        void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    };

    /** One- or two-argument function. Note that this special-cases over
        the opcode to deal with a missing second argument. */
    class IntOneTwoFunctionCallNode : public IntFunctionCallNode {
        uint8_t minor;
     public:
        IntOneTwoFunctionCallNode(uint8_t minor)
            : minor(minor)
            { }
        void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    };

    /** "Mid" function. This function takes two or three arguments, and
        is implemented as one or two binary operations. */
    class IntMidFunctionCallNode : public IntFunctionCallNode {
     public:
        void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    };

    /** "StrCase" function. This function just changes the compilation
        environment for its argument. */
    class IntStrCaseFunctionCallNode : public IntFunctionCallNode {
     public:
        void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    };

    /** "Key" function. The first argument of this function is a
        keymap, not an expression. */
    class IntKeyFunctionCallNode : public IntFunctionCallNode {
     public:
        void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    };

    /** "Eval" function. This unary function evaluates its string
        argument as an expression. */
    class IntEvalFunctionCallNode : public IntFunctionCallNode {
     public:
        void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    };

    // /** "Array" function. Takes any number of arguments and turns them
    //     into an array. */
    // class IntArrayFunctionCallNode : public IntFunctionCallNode {
    //  public:
    //     void compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc);
    // };
}

/************************* IntIfFunctionCallNode *************************/

void
IntIfFunctionCallNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    IntBytecodeObject::label_t ift = bco.makeLabel();
    IntBytecodeObject::label_t iff = bco.makeLabel();
    IntBytecodeObject::label_t end = bco.makeLabel();
    args[0]->compileCondition(bco, cc, ift, iff);
    bco.addLabel(ift);
    args[1]->compileValue(bco, cc);
    bco.addJump(IntOpcode::jAlways, end);
    bco.addLabel(iff);
    if (args.size() > 2)
        args[2]->compileValue(bco, cc);
    else
        bco.addPushLiteral(0);
    bco.addLabel(end);
}

void
IntFoldFunctionCallNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    if (args.size() == 0) {
        // Special case (does not appear)
        bco.addPushLiteral(0);
    } else {
        // Handle case sensitivity
        uint8_t m = minor;
        if ((minor == biMin || minor == biMax) && (cc.compilation_flags & IntCompilationContext::CaseBlind) != 0)
            ++m;

        // First arg
        args[0]->compileValue(bco, cc);

        // Remaining args
        for (uint32_t i = 1; i < args.size(); ++i) {
            args[i]->compileValue(bco, cc);
            bco.addInstruction(IntOpcode::maBinary, m, 0);
        }

        // Type check args. This makes sure that "BitAnd('foo')" fails.
        if (args.size() == 1 && (minor == biBitAnd || minor == biBitOr || minor == biBitXor)) {
            bco.addInstruction(IntOpcode::maPush,   IntOpcode::sInteger, 0);
            bco.addInstruction(IntOpcode::maBinary, biBitOr,             0);
        }
    }
}

void
IntBuiltinFunctionNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    if (args.size() == 1) {
        args[0]->compileValue(bco, cc);
        bco.addInstruction(IntOpcode::maUnary, minor, 0);
    } else if (args.size() == 2) {
        args[0]->compileValue(bco, cc);
        args[1]->compileValue(bco, cc);
        bco.addInstruction(IntOpcode::maBinary, minor, 0);
    } else if (args.size() == 0) {
        bco.addInstruction(IntOpcode::maSpecial, minor, 0);
    } else {
        ASSERT(0);
    }
}

void
IntCaseFunctionNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    ASSERT(args.size() == 2);
    args[0]->compileValue(bco, cc);
    args[1]->compileValue(bco, cc);
    bco.addInstruction(IntOpcode::maBinary,
                       (cc.compilation_flags & IntCompilationContext::CaseBlind) ? minor+1 : minor,
                       0);
}

/** Helper for IntFindFunctionCallNode/IntCountFunctionCallNode: compile "array" expression.
    This is given explicitly for the generic version, and a hardcoded global variable for
    the specific ones.
    \param bco   Output bytecode object
    \param cc    Compilation context
    \param which Which flavour of Find/Count we're compiling (FC_xxx)
    \param args  Arguments */
static void
compileFCArray(IntBytecodeObject& bco, const IntCompilationContext& cc, uint8_t which, ptr_vector<IntExprNode>& args)
{
    switch (which) {
     case FC_Generic:
        args[0]->compileValue(bco, cc);
        break;
     case FC_Ship:
        bco.addInstruction(IntOpcode::maPush, IntOpcode::sNamedShared, bco.addName("SHIP"));
        break;
     case FC_Planet:
        bco.addInstruction(IntOpcode::maPush, IntOpcode::sNamedShared, bco.addName("PLANET"));
        break;
     default:
        throw IntError::internalError("Not implemented");
    }
}

/** Helper for IntFindFunctionCallNode/IntCountFunctionCallNode: compile "condition" expression.
    This is the second arg for the generic version, the first for the specific ones.
    \param bco   Output bytecode object
    \param cc    Compilation context
    \param which Which flavour of Find/Count we're compiling (FC_xxx)
    \param args  Arguments */
static void
compileFCCondition(IntBytecodeObject& bco, const IntCompilationContext& cc,
                   IntBytecodeObject::label_t ift, IntBytecodeObject::label_t iff,
                   uint8_t which, ptr_vector<IntExprNode>& args)
{
    switch (which) {
     case FC_Generic:
        args[1]->compileCondition(bco, cc, ift, iff);
        break;
     case FC_Ship:
     case FC_Planet:
        args[0]->compileCondition(bco, cc, ift, iff);
        break;
     default:
        throw IntError::internalError("Not implemented");
    }
}

/** Helper for IntFindFunctionCallNode: compile "value" expression.
    This is given explicitly for the generic version, and a hardcoded global variable for
    the specific ones.
    \param bco   Output bytecode object
    \param cc    Compilation context
    \param which Which flavour of Find we're compiling (FC_xxx)
    \param args  Arguments */
static void
compileFCValue(IntBytecodeObject& bco, const IntCompilationContext& cc, uint8_t which, ptr_vector<IntExprNode>& args)
{
    switch (which) {
     case FC_Generic:
        args[2]->compileValue(bco, cc);
        break;
     case FC_Ship:
     case FC_Planet:
        bco.addInstruction(IntOpcode::maPush, IntOpcode::sNamedVariable, bco.addName("ID"));
        break;
     default:
        throw IntError::internalError("Not implemented");
    }
}

void
IntFindFunctionCallNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    //     <array>
    //     firstindex
    //     jfep 2F
    // 1H: <expr>
    //     jtp 3F
    //     nextindex
    //     jt 1B
    // 2H: pushe
    //     j 4F
    // 3H: <result>
    //     endindex
    // 4H:
    IntBytecodeObject::label_t loop     = bco.makeLabel();
    IntBytecodeObject::label_t ift      = bco.makeLabel();
    IntBytecodeObject::label_t iff      = bco.makeLabel();
    IntBytecodeObject::label_t notfound = bco.makeLabel();
    IntBytecodeObject::label_t end      = bco.makeLabel();

    IntCompilationContext ncc = cc;
    ncc.compilation_flags &= ~IntCompilationContext::LocalContext;

    compileFCArray(bco, ncc, which, args);
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialFirstIndex, 0);
    bco.addJump(IntOpcode::jIfFalse | IntOpcode::jIfEmpty | IntOpcode::jPopAlways, notfound);
    bco.addLabel(loop);
    compileFCCondition(bco, ncc, ift, iff, which, args);
    bco.addLabel(iff);
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialNextIndex, 0);
    bco.addJump(IntOpcode::jIfTrue | IntOpcode::jPopAlways, loop);
    bco.addLabel(notfound);
    bco.addPushLiteral(0);
    bco.addJump(IntOpcode::jAlways, end);
    bco.addLabel(ift);
    compileFCValue(bco, ncc, which, args);
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialEndIndex, 0);
    bco.addLabel(end);
}

void
IntCountFunctionCallNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    //     pushint 0
    //     <array>
    //     firstindex
    //     jfep 3F
    // 1H: <expr>
    //     jfep 2F
    //     pushint 1
    //     badd
    // 2H: nextindex
    //     jtp 1B
    // 3H:
    IntBytecodeObject::label_t loop     = bco.makeLabel();
    IntBytecodeObject::label_t end      = bco.makeLabel();
    IntBytecodeObject::label_t ift      = bco.makeLabel();
    IntBytecodeObject::label_t iff      = bco.makeLabel();

    IntCompilationContext ncc = cc;
    ncc.compilation_flags &= ~IntCompilationContext::LocalContext;

    bco.addInstruction(IntOpcode::maPush, IntOpcode::sInteger, 0);
    compileFCArray(bco, cc, which, args);
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialFirstIndex, 0);
    bco.addJump(IntOpcode::jIfFalse | IntOpcode::jIfEmpty | IntOpcode::jPopAlways, end);
    bco.addLabel(loop);
    if (which != FC_Generic || args.size() > 1) {
        compileFCCondition(bco, ncc, ift, iff, which, args);
        bco.addLabel(ift);
    }
    bco.addInstruction(IntOpcode::maUnary, unInc, 0);
    bco.addLabel(iff);
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialNextIndex, 0);
    bco.addJump(IntOpcode::jIfTrue | IntOpcode::jPopAlways, loop);
    bco.addLabel(end);
}

void
IntOneTwoFunctionCallNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    args[0]->compileValue(bco, cc);
    if (args.size() < 2) {
        // Special case for unary Str():
        if (minor == biStr) {
            bco.addInstruction(IntOpcode::maUnary, unStr, 0);
            return;
        }

        // Others have a second argument
        if (minor == biStrMult) {
            // String(...). Does it make more sense to generate 'pushint 32, uchr'?
            IntStringValue sv(" ");
            bco.addPushLiteral(&sv);
        } else {
            // ATan, Dim
            bco.addInstruction(IntOpcode::maPush, IntOpcode::sInteger, 1);
        }
    } else {
        // Second argument
        args[1]->compileValue(bco, cc);
    }
    bco.addInstruction(IntOpcode::maBinary, minor, 0);
}

void
IntMidFunctionCallNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    args[0]->compileValue(bco, cc);
    args[1]->compileValue(bco, cc);
    bco.addInstruction(IntOpcode::maBinary, biLCut, 0);
    if (args.size() > 2) {
        args[2]->compileValue(bco, cc);
        bco.addInstruction(IntOpcode::maBinary, biRCut, 0);
    }
}

void
IntStrCaseFunctionCallNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    IntCompilationContext ncc = cc;
    ncc.compilation_flags &= ~IntCompilationContext::CaseBlind;
    args[0]->compileValue(bco, ncc);
}

void
IntKeyFunctionCallNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    IntIdentifierNode* id = dynamic_cast<IntIdentifierNode*>(args[0]);
    if (!id)
        throw IntError::typeError(IntError::tyKeymap);

    // Push keymap
    IntStringValue keymap(id->getIdentifier());
    bco.addPushLiteral(&keymap);
    bco.addInstruction(IntOpcode::maUnary, unKeyLookup, 0);

    // Evaluate key
    args[1]->compileValue(bco, cc);

    bco.addInstruction(IntOpcode::maBinary, biKeyFind, 0);
}

void
IntEvalFunctionCallNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
{
    args[0]->compileValue(bco, cc);
    bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialEvalExpr, 0);
}

// void
// IntArrayFunctionCallNode::compileValue(IntBytecodeObject& bco, const IntCompilationContext& cc)
// {
//     for (uint32_t i = 0; i < args.size(); ++i)
//         args[i]->compileValue(bco, cc);
//     bco.addInstruction(IntOpcode::maSpecial, IntOpcode::miSpecialMakeList, args.size());
// }

/*************************** External Interface **************************/

static IntFunctionCallNode*
makeIf(const TBuiltinFunctionDescriptor&)
{
    return new IntIfFunctionCallNode();
}

static IntFunctionCallNode*
makeBuiltin(const TBuiltinFunctionDescriptor& desc)
{
    return new IntBuiltinFunctionNode(desc.generator_arg);
}

static IntFunctionCallNode*
makeCase(const TBuiltinFunctionDescriptor& desc)
{
    return new IntCaseFunctionNode(desc.generator_arg);
}

static IntFunctionCallNode*
makeFold(const TBuiltinFunctionDescriptor& desc)
{
    return new IntFoldFunctionCallNode(desc.generator_arg);
}

static IntFunctionCallNode*
makeFind(const TBuiltinFunctionDescriptor& desc)
{
    return new IntFindFunctionCallNode(desc.generator_arg);
}

static IntFunctionCallNode*
makeCount(const TBuiltinFunctionDescriptor& desc)
{
    return new IntCountFunctionCallNode(desc.generator_arg);
}

static IntFunctionCallNode*
makeOneTwo(const TBuiltinFunctionDescriptor& desc)
{
    return new IntOneTwoFunctionCallNode(desc.generator_arg);
}

static IntFunctionCallNode*
makeMid(const TBuiltinFunctionDescriptor&)
{
    return new IntMidFunctionCallNode();
}

static IntFunctionCallNode*
makeStrCase(const TBuiltinFunctionDescriptor&)
{
    return new IntStrCaseFunctionCallNode();
}

static IntFunctionCallNode*
makeKey(const TBuiltinFunctionDescriptor&)
{
    return new IntKeyFunctionCallNode();
}

static IntFunctionCallNode*
makeEval(const TBuiltinFunctionDescriptor&)
{
    return new IntEvalFunctionCallNode();
}

// static IntFunctionCallNode*
// makeArray(const TBuiltinFunctionDescriptor&)
// {
//     return new IntArrayFunctionCallNode();
// }

static const TBuiltinFunctionDescriptor builtin_functions[] = {
    { "ABS",         1, 1,       makeBuiltin, unAbs },
    // { "ARRAY",    0, INT_MAX, makeArray,   0 },
    { "ASC",         1, 1,       makeBuiltin, unAsc },
    { "ATAN",        1, 2,       makeOneTwo,  biATan },
    { "ATOM",        1, 1,       makeBuiltin, unAtom },
    { "ATOMSTR",     1, 1,       makeBuiltin, unAtomStr },
    { "BITAND",      1, INT_MAX, makeFold,    biBitAnd },
    { "BITNOT",      1, 1,       makeBuiltin, unBitNot },
    { "BITOR",       1, INT_MAX, makeFold,    biBitOr },
    { "BITXOR",      1, INT_MAX, makeFold,    biBitXor },
    { "CC$TRACE",    1, 1,       makeBuiltin, unTrace },
    { "CHR",         1, 1,       makeBuiltin, unChr },
    { "CHR$",        1, 1,       makeBuiltin, unChr },
    { "COS",         1, 1,       makeBuiltin, unCos },
    { "COUNT",       1, 2,       makeCount,   FC_Generic },
    { "COUNTPLANETS",1, 1,       makeCount,   FC_Planet },
    { "COUNTSHIPS",  1, 1,       makeCount,   FC_Ship },
    { "DIM",         1, 2,       makeOneTwo,  biArrayDim },
    { "EVAL",        1, 1,       makeEval,    0 },
    { "EXP",         1, 1,       makeBuiltin, unExp },
    { "FIND",        3, 3,       makeFind,    FC_Generic },
    { "FINDPLANET",  1, 1,       makeFind,    FC_Planet },
    { "FINDSHIP",    1, 1,       makeFind,    FC_Ship },
    { "FIRST",       2, 2,       makeCase,    biFirstStr },
    { "IF",          2, 3,       makeIf,      0 },
    { "INSTR",       2, 2,       makeCase,    biFindStr },
    { "INT",         1, 1,       makeBuiltin, unTrunc },
    { "ISARRAY",     1, 1,       makeBuiltin, unIsArray },
    { "ISEMPTY",     1, 1,       makeBuiltin, unIsEmpty },
    { "ISNUM",       1, 1,       makeBuiltin, unIsNum },
    { "ISSTRING",    1, 1,       makeBuiltin, unIsString },
    { "KEY",         2, 2,       makeKey,     0 },
    { "LEFT",        2, 2,       makeBuiltin, biRCut },
    { "LEN",         1, 1,       makeBuiltin, unLength },
    { "LOG",         1, 1,       makeBuiltin, unLog },
    { "LTRIM",       1, 1,       makeBuiltin, unLTrim },
    { "MAX",         1, INT_MAX, makeFold,    biMax },
    { "MID",         2, 3,       makeMid,     0 },
    { "MIN",         1, INT_MAX, makeFold,    biMin },
    { "NEWHASH",     0, 0,       makeBuiltin, IntOpcode::miSpecialNewHash },
    { "REST",        2, 2,       makeCase,    biRestStr },
    { "RIGHT",       2, 2,       makeBuiltin, biEndCut },
    { "ROUND",       1, 1,       makeBuiltin, unRound },
    { "RTRIM",       1, 1,       makeBuiltin, unRTrim },
    { "SIN",         1, 1,       makeBuiltin, unSin },
    { "SQR",         1, 1,       makeBuiltin, unSqrt },
    { "SQRT",        1, 1,       makeBuiltin, unSqrt },
    { "STR",         1, 2,       makeOneTwo,  biStr },
    { "STRCASE",     1, 1,       makeStrCase, 0 },
    { "STRING",      1, 2,       makeOneTwo,  biStrMult },
    { "STRING$",     1, 2,       makeOneTwo,  biStrMult },
    { "TAN",         1, 1,       makeBuiltin, unTan },
    { "TRIM",        1, 1,       makeBuiltin, unLRTrim },
    { "VAL",         1, 1,       makeBuiltin, unVal },
    { "Z",           1, 1,       makeBuiltin, unZap },
    { "ZAP",         1, 1,       makeBuiltin, unZap },
};

/** Look up descriptor for a builtin function. Builtin functions are directly
    encoded into the bytecode, and can thus not be redefined by the user. */
const TBuiltinFunctionDescriptor*
lookupBuiltinFunction(string_t name)
{
    /* Quick & Dirty. */
    for (std::size_t i = 0; i < countof(builtin_functions); ++i)
        if (name == builtin_functions[i].name)
            return &builtin_functions[i];
    return 0;
}
