/**
  *  \file int/value.cc
  *  \brief Interpreter: Values
  *
  *  This implements the basic user-visible types for the interpreter,
  *  for in-memory storage and serialisation.
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <math.h>
#include "int/value.h"
#include "cpluslib/format.h"
#include "io/streams.h"
#include "io/io.h"
#include "util/except.h"
#include "util/nls.h"
#include "defs.h"
#include "int/contextio.h"
#include "int/context.h"
#include "int/vmio.h"
#include "util/recode.h"
#include "int/compound.h"
#include "int/hash.h"
#include "int/array.h"

/// Serialisation information for IntTagNode.
const TStructureParser IntTagNode::parser[] = {
    { T_INT16, 1, offsetof(IntTagNode, tag), 0, 0 },
    { T_INT32, 1, offsetof(IntTagNode, value), 0, 0 },
    { T_END, 0, 0, 0, 0 }
};

/******************************** IntValue *******************************/

static IntValue*
failInvalidValue(Stream& aux)
{
    throw FileFormatException(aux, _("Invalid value in data segment; file probably written by newer version of PCC"));
}

IntValue::~IntValue()
{ }

IntValue*
IntValue::load(const IntTagNode& sv, Stream& aux)
{
    switch (sv.tag) {
     case IntTagNode::Tag_Empty:
        return 0;

     case IntTagNode::Tag_Integer:
     case IntTagNode::Tag_Boolean:
        return new IntIntValue(sv, aux);

     case IntTagNode::Tag_String:
     case IntTagNode::Tag_LongString:
        return new IntStringValue(sv, aux);

     case IntTagNode::Tag_32bitFP:
     case IntTagNode::Tag_FPZero:
        return new IntFloatValue(sv, aux);

     case IntTagNode::Tag_Blob:
        return new IntBlobValue(sv, aux);

     case IntTagNode::Tag_FileHandle:
        return new IntFileValue(sv, aux);

     case IntTagNode::Tag_BCO:
        if (IntVMLoadContext* vlc = IntVMLoadContext::getCurrentInstance()) {
            return new IntSubroutineValue(vlc->getBCO(sv.value));
        } else {
            return failInvalidValue(aux);
        }

     case IntTagNode::Tag_Array:
        if (IntVMLoadContext* vlc = IntVMLoadContext::getCurrentInstance()) {
            return new IntArray(vlc->getArray(sv.value));
        } else {
            return failInvalidValue(aux);
        }

     case IntTagNode::Tag_Hash:
        if (IntVMLoadContext* vlc = IntVMLoadContext::getCurrentInstance()) {
            return new IntHash(vlc->getHash(sv.value));
        } else {
            return failInvalidValue(aux);
        }

     default:
        if ((sv.tag & 0x00FF) == 0) {
            IntContext* cv = loadContext(sv, aux);
            if (!cv)
                return failInvalidValue(aux);
            return cv;
        } else {
            return new IntFloatValue(sv, aux);
        }
    }
}

IntValue*
IntValue::cloneOf(IntValue* p)
{
    if (p) {
        return p->clone();
    } else {
        return 0;
    }
}

string_t
IntValue::toStringOf(IntValue* p, bool readable)
{
    if (p) {
        return p->toString(readable);
    } else {
        if (readable) {
            return "Z(0)";
        } else {
            return string_t();
        }
    }
}

/****************************** IntIntValue ******************************/

/** Construct from literal.
    \param value   Value
    \param is_bool true to make a boolean, false to make an integer. Those behave
                   mostily identical, but stringify differently. */
IntIntValue::IntIntValue(int32 value, bool is_bool)
    : value(value),
      is_bool(is_bool)
{ }

/** Construct from serialized form.
    \param sv      Tag read from file
    \param aux     Auxiliary data field */
IntIntValue::IntIntValue(const IntTagNode& sv, Stream& /*aux*/)
    : value(sv.value),
      is_bool(sv.tag == sv.Tag_Boolean)
{ }

IntIntValue::~IntIntValue()
{ }

string_t
IntIntValue::toString(bool readable)
{
    if (is_bool) {
        if (readable)
            return value ? "True" : "False";
        else
            return value ? "YES" : "NO";
    } else {
        return format("%d", value);
    }
}

void
IntIntValue::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = is_bool ? sv.Tag_Boolean : sv.Tag_Integer;
    sv.value = value;
}

IntIntValue*
IntIntValue::clone()
{
    return new IntIntValue(*this);
}

/***************************** IntStringValue ****************************/

/** Construct from literal.
    \param value string literal */
IntStringValue::IntStringValue(string_t value)
    : value(value)
{ }

/** Construct from serialized form.
    \param sv      Tag read from file
    \param aux     Auxiliary data field */
IntStringValue::IntStringValue(const IntTagNode& sv, Stream& aux)
    : value()
{
    if (sv.tag == sv.Tag_String) {
        // Short String
        if (sv.value != 0) {
            getPascalStringT(aux, value);
        }
    } else {
        // Long String
        getString(aux, value, sv.value);
    }
    value = convertGameToUtf8(value);
}

IntStringValue::~IntStringValue()
{ }

string_t
IntStringValue::toString(bool readable)
{
    if (readable) {
        return quoteString(value);
    } else {
        return value;
    }
}

void
IntStringValue::store(IntTagNode& sv, Stream& aux)
{
    string_t converted = convertUtf8ToGame(value);
    if (converted.size() == 0) {
        // Empty string. Use Short String format, which is supported by PCC 1.0.8+.
        sv.tag   = sv.Tag_String;
        sv.value = 0;
    } else if (converted.size() <= 255 && IntVMSaveContext::getCurrentInstance() == 0) {
        // Use Short String format, which is supported by PCC 1.0.8+.
        // When we're saving a VM file (that is, when we have a IntVMSaveContext
        // instance), we don't have to care for backward compatibility, so let's
        // save a few bytes by not using short string format there.
        sv.tag   = sv.Tag_String;
        sv.value = 0x11111111;
        storePascalStringT(aux, converted);
    } else {
        // Use Long String format, PCC 1.0.18+.
        sv.tag   = sv.Tag_LongString;
        sv.value = converted.size();
        aux.writeT(converted.data(), converted.size());
    }
}

IntStringValue*
IntStringValue::clone()
{
    return new IntStringValue(*this);
}

/***************************** IntFloatValue *****************************/

/** Convert REAL (real48) into double.
    \param sv contains a REAL */
static double
getFP48(const IntTagNode& sv)
{
    // REAL format:                                      FLOAT format:
    //   8 bit exponent "e"                               23 bit mantissa "m"
    //  39 bit mantissa "m"                                8 bit exponent "e"
    //   1 bit sign "s"                                    1 bit sign "s"
    // Value is e=0 => 0.0                               e=0   => (-1)^s * (2^-126) * (0.m)
    //          e>0 => (-1)^s * 2^(e-129) * (1.m)        e=255 => Inf, NaN
    //                                                   sonst => (-1)^s * 2^(e-127) * (1.m)

    int e = sv.tag & 255;
    if (e == 0) {
        return 0.0;
    } else {
        int s    = (sv.value & 0x80000000U) ? -1 : +1;
        /* Generate the mantissa 1.m times 2**39 */
        double m = (sv.value & 0x7FFFFFFF) * 256.0 + (sv.tag >> 8) + 549755813888.0 /* 2**39 */;
        return s * ldexp(m, e - (129 + 39));
    }

#if 0
    // The following converts the REAL into a FLOAT by manipulating the bit
    // representation. It is less precise, as it generates only 23 mantissa
    // bits (i.e. loses 16 bits).
    uint32 e = sv.tag & 255;           // REAL exponent
    uint32 m = sv.value & 0x7FFFFFFF;  // REAL mantissa, 31 bits (ignoring 8 LSBs)
    uint32 s = sv.value >> 31;         // REAL sign

    // Convert to FLOAT mantissa. We have 31 bits, want 23, so shift by 8.
    m >>= 8;

    if (e <= 2) {
        // Result will be FLOAT zero
        return 0.0;
    } else {
        // Assemble new value
        uint32 nv = m | ((e - 2) << 23) | (s << 31);
        float fv;
        memcpy(&fv, &nv, sizeof(nv));
        return fv;
    }
#endif
}

/** Store value in REAL (real48) format. REAL cannot store infinities, and has a smaller
    range than usual doubles. Those are stored as REAL-max. Underflows turn into REAL
    zeroes. REAL cannot store NaNs; those are converted into Tag_Empty. Although this
    does change the value in an observable way, it's probably the best we can do. The
    interpreter should try to avoid such values.
    \param value [in] Value
    \param sv [out] Stored value */
static void
storeFP48(double value, IntTagNode& sv)
{
    // REAL format:
    //   8 bit exponent "e"
    //  39 bit mantissa "m"
    //   1 bit sign "s"
    // Value is e=0 => 0.0
    //          e>0 => (-1)^s * 2^(e-129) * (1.m)
    // Maximum REAL is 1.111111111111111111111111111111111111111 * 2^126
    //            i.e. 170141183460314489226776631181521715200
    //            i.e. 1.7e+38
    // We cannot store infinities and NaNs.
    // - convert infinities and overflows to max REAL
    // - convert underflows to 0.0
    // - convert NaNs to EMPTY (!)
#ifdef HAVE_ISNAN
    if (isnan(value)) {
        sv.tag   = IntTagNode::Tag_Empty;
        sv.value = 0;
        return;
    }
#endif

    // Find out sign
    uint32_t sign;
    if (value < 0) {
        sign  = 0x80000000U;
        value = -value;
    } else {
        sign = 0;
    }

    // First check for infinities or large values
    if (value > 1.0E+39) {
        sv.tag   = 0xFFFF;
        sv.value = 0x7FFFFFFF | sign;
    } else {
        // Reasonable value. Split into exponent and mantissa.
        int exp;
        double mant = frexp(value, &exp);
        exp -= 1;               // mant will be 0.1xxxxxxxxxxxxxx, we want 1.xxxxxxxxxx
        exp += 129;             // adjust for storage
        if (exp <= 0) {
            // Zero or underflow
            sv.tag   = 0;
            sv.value = 0;
        } else if (exp > 255) {
            // Overflow
            sv.tag   = 0xFFFF;
            sv.value = 0x7FFFFFFF | sign;
        } else {
            // It's a regular value. Extract mantissa bits.
            mant -= 0.5;           // 0.0yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyxxxxxxxx
            mant *= 4294967296.0;  // yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy.xxxxxxxx
            int32 bits1 = int32(mant);
            mant -= bits1;         // 0.xxxxxxxx
            mant *= 256.0;
            int32 bits2 = int32(mant);
            // Build result
            sv.tag   = exp | (256*bits2);
            sv.value = bits1 | sign;
        }
    }
}

/** Construct from literal.
    \param value literal */
IntFloatValue::IntFloatValue(double value)
    : value(value)
{ }

/** Construct from serialized form.
    \param sv      Tag read from file
    \param aux     Auxiliary data field */
IntFloatValue::IntFloatValue(const IntTagNode& sv, Stream& /*aux*/)
{
    using std::memcpy;
    if (sv.tag == sv.Tag_FPZero) {
        // Zero
        value = 0.0;
    } else if (sv.tag == sv.Tag_32bitFP) {
        // 32-bit FP
        // FIXME: portability. We now assume that floats have the same endianness as int32s.
        float tmp;
        CTASSERT(sizeof(tmp) == sizeof(sv.value));
        memcpy(&tmp, &sv.value, sizeof(sv.value));
        value = tmp;
    } else {
        // 48-bit FP
        value = getFP48(sv);
    }
}

IntFloatValue::~IntFloatValue()
{ }

string_t
IntFloatValue::toString(bool /*readable*/)
{
    string_t result = format("%f", value);
    if (result.find('.') != result.npos && result.find_first_of("Ee") == result.npos) {
        string_t::size_type cut = result.size();
        while (cut > 0 && result[cut-1] == '0')
            --cut;
        if (cut > 0 && result[cut-1] == '.')
            --cut;
        result.erase(cut);
    }
    return result;
}

void
IntFloatValue::store(IntTagNode& sv, Stream& /*aux*/)
{
    using std::memcpy;
    if (value == 0.0) {
        sv.tag   = sv.Tag_FPZero;
        sv.value = 0;
    } else {
        storeFP48(value, sv);
    }
}

IntFloatValue*
IntFloatValue::clone()
{
    return new IntFloatValue(*this);
}

/****************************** IntBlobValue *****************************/

IntBlobValue::IntBlobValue()
{ }

IntBlobValue::IntBlobValue(const IntTagNode& sv, Stream& aux)
    : value(sv.value, char(0))
{
    if (sv.value != 0)
        aux.readT(&value.front(), sv.value);
}

IntBlobValue::~IntBlobValue()
{ }

string_t
IntBlobValue::toString(bool /*readable*/)
{
    return format("#<data:%d>", value.size());
}

void
IntBlobValue::store(IntTagNode& sv, Stream& aux)
{
    sv.tag   = sv.Tag_Blob;
    sv.value = value.size();
    if (value.size())
        aux.writeT(&value.front(), value.size());
}

IntBlobValue*
IntBlobValue::clone()
{
    return new IntBlobValue(*this);
}

/****************************** IntFileValue *****************************/

IntFileValue::IntFileValue(int32_t file_nr)
    : file_nr(file_nr)
{ }

IntFileValue::IntFileValue(const IntTagNode& sv, Stream& /*aux*/)
    : file_nr(sv.value)
{ }

IntFileValue::~IntFileValue()
{ }

string_t
IntFileValue::toString(bool /*readable*/)
{
    return format("#%d", file_nr);
}

void
IntFileValue::store(IntTagNode& sv, Stream& /*aux*/)
{
    sv.tag   = IntTagNode::Tag_FileHandle;
    sv.value = file_nr;
}

IntFileValue*
IntFileValue::clone()
{
    return new IntFileValue(file_nr);
}

/******************************** Globals ********************************/

/** Make a tristate-boolean value from integer.
    \param value Negative for empty, zero for false, positive for true */
IntValue*
makeBoolValue(int value)
{
    if (value < 0)
        return 0;
    else
        return new IntIntValue(value != 0, true);
}

/** Make integer value. Convenience function to save code.
    \param value Value to create */
IntValue*
makeIntValue(int32_t value)
{
    return new IntIntValue(value, false);
}

/** Make float value. Convenience function to save code.
    \param value Value to create */
IntValue*
makeFloatValue(double value)
{
    return new IntFloatValue(value);
}

/** Make string value. Convenience function to save code.
    \param str Value to create */
IntValue*
makeStringValue(string_t str)
{
    return new IntStringValue(str);
}

/** Make string value. Convenience function to save code.
    \param str Value to create */
IntValue*
makeStringValue(const char* str)
{
    return new IntStringValue(str);
}

/** Get tristate-integer value from value. This is used whenever a value is used
    in a boolean context in a program.
    \retval -1 Input is EMPTY
    \retval 0  Input is False
    \retval +1 Input is True */
int
getBoolValue(IntValue* value)
{
    // FIXME: what should this return for values that are neither int, nor float, nor string?
    // Right now, it makes them EMPTY, but should probably generate an error. However, generating
    // EMPTY has the convenient property of making ubool/unot/uzap/unot2/uisempty operations
    // never fail, which permits a number of useful optimisations.
    if (value == 0)
        return -1;
    else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(value))
        return iv->getValue() != 0;
    else if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(value))
        return fabs(fv->getValue()) > 1.0E-06;
    else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(value))
        return sv->getValue().size() != 0;
    else
        return -1;
}

/** Quote a string. Returns some piece of code that, when read, returns a string \c value. */
string_t
quoteString(const string_t& value)
{
    if (value.find_first_of("\"\\") == value.npos) {
        // No meta-characters, use unquoted double-quote string
        return "\"" + value + "\"";
    } else if (value.find('\'') == value.npos) {
        // Double-quotes or backslashes, but no apostrophes
        return "'" + value + "'";
    } else {
        // Sufficiently complicated, so add quotes.
        string_t output = "\"";
        string_t::size_type i = 0, j;
        while (((j = value.find_first_of("\"\\", i)) != value.npos)) {
            output.append(value, i, j-i);
            output.append(1, '\\');
            output.append(value, j, 1);
            i = j+1;
        }
        output.append(value, i, value.npos);
        output += "\"";
        return output;
    }
}
